from __future__ import division

import numpy as np
from scipy import linalg as la

try:
    import mkl_fft._numpy_fft as fft
except ImportError:
    import numpy.fft as fft

def deconvolve(y, s, length=None, thresh=0.0):
    '''
    Deconvolve an excitation signal from an impulse response

    Parameters
    ----------

    y : ndarray
        The recording
    s : ndarray
        The excitation signal
    length: int, optional
        the length of the impulse response to deconvolve
    thresh : float, optional
        ignore frequency bins with power lower than this
    '''

    # FFT length including zero padding
    n = y.shape[0] + s.shape[0] - 1

    # let FFT size be even for convenience
    if n % 2 != 0:
        n += 1

    # when unknown, pick the filter size as size of test signal
    if length is None:
        length = n

    # Forward transforms
    Y  = fft.rfft(np.array(y, dtype=np.float32), n=n) / np.sqrt(n)
    S = fft.rfft(np.array(s, dtype=np.float32), n=n) / np.sqrt(n)

    # Only do the division where S is large enough
    H = np.zeros(*Y.shape, dtype=Y.dtype)
    I = np.where(np.abs(S) > thresh)
    H[I] = Y[I] / S[I]

    # Inverse transform
    h = fft.irfft(H, n=n)

    return h[:length]

def wiener_deconvolve(y, x, length=None, noise_variance=1., let_n_points=15, let_div_base=2):
    '''
    Deconvolve an excitation signal from an impulse response

    We use Wiener filter

    Parameters
    ----------

    y : ndarray
        The recording
    x : ndarray
        The excitation signal
    length: int, optional
        the length of the impulse response to deconvolve
    noise_variance : float, optional
        estimate of the noise variance
    let_n_points: int
        number of points to use in the LET approximation
    let_div_base: float
        the divider used for the LET grid
    '''

    # FFT length including zero padding
    n = y.shape[0] + x.shape[0] - 1

    # let FFT size be even for convenience
    if n % 2 != 0:
        n += 1

    # when unknown, pick the filter size as size of test signal
    if length is None:
        length = n

    # Forward transforms
    Y  = fft.rfft(np.array(y, dtype=np.float32), n=n) / np.sqrt(n)  # recording
    X = fft.rfft(np.array(x, dtype=np.float32), n=n) / np.sqrt(n)   # test signal

    # Squared amplitude of test signal
    X_sqm = np.abs(X)**2

    # approximate SNR
    SNR_hat = np.maximum(1e-7, ((np.linalg.norm(Y)**2 / np.linalg.norm(X)**2) - noise_variance))  / noise_variance
    dividers = let_div_base**np.linspace(-let_n_points/2, let_n_points, let_n_points)
    SNR_grid = SNR_hat / dividers

    # compute candidate points
    G = (X_sqm[:,None] / (X_sqm[:,None] + 1./SNR_grid[None,:])) * Y[:,None]
    H_candidates = G / X[:,None]

    # find the best linear combination of the candidates
    weights = np.linalg.lstsq(G, Y, rcond=None)[0]

    # compute the estimated filter
    H = np.squeeze(np.dot(H_candidates, weights))
    h = fft.irfft(H, n=n)

    return h[:length]




from __future__ import division, print_function
import pyroomacoustics as pra
import numpy as np

try:
    from pyroomacoustics import build_rir
    build_rir_available = True
except:
    print('build_rir not available')
    build_rir_available = False

# tolerance for test success (1%)
tol = 0.01

fdl = 81
fs = 16000

t0 = (2 * fdl + 0.1) / fs
t1 = (3 * fdl - 0.1) / fs
t2 = (4 * fdl + 0.45) / fs
t3 = (5 * fdl + 0.001) / fs
t4 = (6 * fdl + 0.999) / fs

times = np.array(
        [
            [ t0 , t1 + (1 / 40 / 16000), t2, ],
            [ t0, t1 + (10 / fs), 3 * t3, ],
            [ t0, t3, t4, ],
            ],
        )
alphas = np.array(
        [
            [ 1., 0.5, -0.1 ],
            [ 0.5, 0.3, 0.1 ],
            [ 0.3, 2., 0.1 ],
            ],
        )
visibilities = np.array(
        [
            [ 1, 1, 1,],
            [ 1, 1, 1,],
            [ 0, 1, 1,],
            ],
        dtype=np.int32,
        )


def build_rir_wrap(time, alpha, visibility, fs, fdl):

    # fractional delay length
    fdl = pra.constants.get('frac_delay_length')
    fdl2 = (fdl-1) // 2

    # the number of samples needed
    N = int(np.ceil(time.max() * fs) + fdl)

    ir_ref = np.zeros(N)
    ir_cython = np.zeros(N)

    # Try to use the Cython extension
    build_rir.fast_rir_builder(ir_cython, time, alpha, visibility, fs, fdl)

    # fallback to pure Python implemenation
    for i in range(time.shape[0]):
        if visibility[i] == 1:
            time_ip = int(np.round(fs * time[i]))
            time_fp = (fs * time[i]) - time_ip
            ir_ref[time_ip-fdl2:time_ip+fdl2+1] += alpha[i] * pra.fractional_delay(time_fp)

    return ir_ref, ir_cython

def test_build_rir():

    if not build_rir_available:
        return

    for t, a, v in zip(times, alphas, visibilities):
        ir_ref, ir_cython = build_rir_wrap(times[0], alphas[0], visibilities[0], fs, fdl)
        assert np.max(np.abs(ir_ref - ir_cython)) < tol

def test_short():
    ''' Tests that an error is raised if a provided time goes below the zero index '''

    if not build_rir_available:
        return

    N = 100
    fs = 16000
    fdl = 81
    rir = np.zeros(N)

    time = np.array([0.])
    alpha = np.array([1.])
    visibility = np.array([1], dtype=np.int32)

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility, fs, fdl)
        assert False
    except AssertionError:
        print('Ok, short times are caught')



def test_long():
    ''' Tests that an error is raised if a time falls outside the rir array '''

    if not build_rir_available:
        return

    N = 100
    fs = 16000
    fdl = 81
    rir = np.zeros(N)

    time = np.array([(N-1) / fs])
    alpha = np.array([1.])
    visibility = np.array([1], dtype=np.int32)

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility, fs, fdl)
        assert False
    except AssertionError:
        print('Ok, long times are caught')

def test_errors():
    ''' Tests that errors are raised when array lengths differ '''

    if not build_rir_available:
        return

    N = 300
    fs = 16000
    fdl = 81
    rir = np.zeros(N)

    time = np.array([100 / fs, 200 / fs])
    alpha = np.array([1., 1.])
    visibility = np.array([1, 1], dtype=np.int32)

    try:
        build_rir.fast_rir_builder(rir, time, alpha[:1], visibility, fs, fdl)
        assert False
    except:
        print('Ok, alpha error occured')
        pass

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility[:1], fs, fdl)
        assert False
    except:
        print('Ok, visibility error occured')
        pass

    try:
        build_rir.fast_rir_builder(rir, time, alpha, visibility, fs, 80)
        assert False
    except:
        print('Ok, fdl error occured')
        pass


if __name__ == '__main__':

    import matplotlib.pyplot as plt

    for t, a, v in zip(times, alphas, visibilities):
        ir_ref, ir_cython = build_rir_wrap(times[0], alphas[0], visibilities[0], fs, fdl)

        print('Error:', np.max(np.abs(ir_ref - ir_cython)))

        plt.figure()
        plt.plot(ir_ref, label='ref')
        plt.plot(ir_cython, label='cython')
        plt.legend()

    test_short()
    test_long()
    test_errors()

    plt.show()




'''
Example of the basic operations with ``pyroomacoustics.datasets.Dataset``
and ``pyroomacoustics.datasets.Sample`` classes
'''
from pyroomacoustics.datasets import Sample, Dataset

def test_dataset():
    # Prepare a few artificial samples
    samples = [
        {
            'data' : 0.99,
            'metadata' : { 'speaker' : 'alice', 'sex' : 'female', 'age' : 37, 'number' : 'one' },
        },
        {
            'data' : 2.1,
            'metadata' : { 'speaker' : 'alice', 'sex' : 'female', 'age' : 37, 'number' : 'two' },
        },
        {
            'data' : 1.02,
            'metadata' : { 'speaker' : 'bob', 'sex' : 'male', 'age' : 48, 'number' : 'one' },
        },
        {
            'data' : 2.07,
            'metadata' : { 'speaker' : 'bob', 'sex' : 'male', 'age' : 48, 'number' : 'two' },
        },
        {
            'data' : 3.0,
            'metadata' : { 'speaker' : 'bob', 'sex' : 'male', 'age' : 48, 'number' : 'three' },
        },
        {
            'data' : 1.97,
            'metadata' : { 'speaker' : 'charles', 'sex' : 'male', 'age' : 25, 'number' : 'two' },
        },
        ]

    dataset = Dataset()
    for s in samples:
        new_sample = Sample(s['data'], **s['metadata'])
        dataset.add_sample(new_sample)

    assert len(dataset) == len(samples)


    assert dataset.info['speaker']['alice'] == 2
    assert dataset.info['speaker']['bob'] == 3
    assert dataset.info['speaker']['charles'] == 1

    assert dataset.info['sex']['male'] == 4
    assert dataset.info['sex']['female'] == 2

    assert dataset.info['age'][37] == 2
    assert dataset.info['age'][48] == 3
    assert dataset.info['age'][25] == 1

    assert dataset.info['number']['one'] == 2
    assert dataset.info['number']['two'] == 3
    assert dataset.info['number']['three'] == 1

    assert dataset[2].meta.speaker == 'bob'
    assert dataset[2].meta.sex == 'male'
    assert dataset[2].meta.age == 48
    assert dataset[2].meta.number == 'one'

    # We can obtain a new dataset with only male subject
    dataset2 = dataset.filter(sex='male')

    assert len(dataset2) == 4
    assert dataset2[0].meta.sex == 'male'
    assert dataset2[1].meta.sex == 'male'
    assert dataset2[2].meta.sex == 'male'
    assert dataset2[3].meta.sex == 'male'

    # select all samples with age larger than 30
    dataset3 = dataset.filter(age=lambda a : a > 30)
    assert len(dataset3) == 5
    for s in dataset3.samples:
        assert s.meta.age > 30

    # select all samples with a number that contains 't'
    dataset_contains_n = dataset.filter(number=['one','three'], speaker=['alice','bob'])
    assert len(dataset_contains_n) == 3
    for s in dataset_contains_n.samples:
        assert ( (s.meta.number == 'one' or s.meta.number == 'three')
                and (s.meta.speaker == 'alice' or s.meta.speaker == 'bob') )

if __name__ == '__main__':
    test_dataset()



# Test of Wall constructor
# Copyright (C) 2019  Robin, Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
import pyroomacoustics as pra

eps = 1e-6

# The vertices of the wall are assumed to turn counter-clockwise around the
# normal of the wall

# A very simple wall
walls = [
        {
            'corners' : np.array([
                [0., 1., 1., 0.],
                [0., 0., 1., 1.],
                [0., 0., 0., 0.],
                ]),
            'area' : 1,
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        {
            'corners' : np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
                ]),
            'area' : 3.4641016151377557,  # this is an equilateral triangle with side sqrt(8)
            'absorption' : [0.2],
            'scattering' : [0.1],
            },
        ]



def test_wall_3d_construct_0():
    ''' Tests construction of a wall '''
    w_info = walls[0]
    wall = pra.wall_factory(w_info['corners'], [0.2], [0.1])

    return wall

def test_wall_3d_construct_1():
    ''' Tests construction of a wall '''
    w_info = walls[1]
    wall = pra.wall_factory(w_info['corners'], [0.2], [0.1])

    return wall

def test_wall_3d_area_0():
    ''' Tests the area computation '''
    w_info = walls[0]
    wall = pra.wall_factory(w_info['corners'], w_info['absorption'], w_info['scattering'])
    err = abs(wall.area() - w_info['area'])
    assert err < 1, 'The error is {}'.format(err)

def test_wall_3d_area_1():
    ''' Tests the area computation '''
    w_info = walls[1]
    wall = pra.wall_factory(w_info['corners'], w_info['absorption'], w_info['scattering'])
    err = abs(wall.area() - w_info['area'])
    assert err < 1, 'The error is {}'.format(err)

def test_wall_3d_normal_0():
    ''' Tests direction of normal wrt to point arrangement '''
    w_info = walls[0]
    wall1 = pra.wall_factory(w_info['corners'], [0.2], [0.1])
    # the same wall with normal pointing the other way
    wall2 = pra.wall_factory(w_info['corners'][:,::-1], [0.2], [0.1])

    err = np.linalg.norm(wall1.normal + wall2.normal)
    assert err < eps, 'The error is {}'.format(err)

def test_wall_3d_normal_1():
    ''' Tests direction of normal wrt to point arrangement '''
    w_info = walls[1]
    wall1 = pra.wall_factory(w_info['corners'], [0.2], [0.1])
    # the same wall with normal pointing the other way
    wall2 = pra.wall_factory(w_info['corners'][:,::-1], [0.2], [0.1])

    err = np.linalg.norm(wall1.normal + wall2.normal)
    assert err < eps, 'The error is {}'.format(err)

if __name__ == '__main__':

        wall0 = test_wall_3d_construct_0()
        test_wall_3d_normal_0()
        test_wall_3d_area_0()
        wall1 = test_wall_3d_construct_1()
        test_wall_3d_normal_1()
        test_wall_3d_area_1()




import numpy as np
from scipy.signal import fftconvolve
import pyroomacoustics as pra

def test_correlate():
    
    N = [100, 200, 50, 37]
    M = [47, 82, 151, 893]

    for n, m in zip(N,M):

        x = np.random.randn(n)
        y = np.random.randn(m)

        assert np.allclose(pra.correlate(x,y), np.correlate(x, y, mode='full'))

def test_tdoa_delay_int():

    N = [100, 200, 50, 1000]
    M = [47, 37, 12, 128]
    delays = [27, 4, 10, 347]

    for n,m,tau in zip(N,M,delays):

        pulse = np.random.randn(m)
        x1 = np.zeros(n)
        x1[:m] = pulse
        x2 = np.zeros(n)
        x2[tau:tau+m] = pulse

        d1 = pra.tdoa(x1, x2)
        d2 = pra.tdoa(x2, x1)

        assert int(d1) == -tau
        assert int(d2) == tau

def test_tdoa_delay_frac_phat():

    N = [14, 200, 70, 40, 28]
    M = [3, 78, 12, 10, 12]

    # important to start at zero and only use positive
    # delays due to fractional_delay_filter_bank function
    # subtracting minimum delay
    delays = [0.0, 0.3, 3.5, 10.1, 5.7]

    zero_delay_filter = pra.fractional_delay(0.)
    frac_filters = pra.fractional_delay_filter_bank(delays)

    # test without explicit sampling frequency
    for n,m,tau,fil in zip(N,M,delays,frac_filters):

        pulse = np.random.randn(m)
        signal = np.zeros(n)
        signal[:m] = pulse
        x1 = fftconvolve(zero_delay_filter, signal)
        x2 = fftconvolve(signal, fil)

        d1 = pra.tdoa(x1, x2, interp=20, phat=True)
        d2 = pra.tdoa(x2, x1, interp=20, phat=True)

        assert np.allclose(d1, -tau)
        assert np.allclose(d2, tau)

        fs = 100
        d3 = pra.tdoa(x1, x2, interp=20, phat=True, fs=fs)
        d4 = pra.tdoa(x2, x1, interp=20, phat=True, fs=fs)

        assert np.allclose(d3, -tau / fs)
        assert np.allclose(d4, tau / fs)





if __name__ == '__main__':

    test_correlate()
    test_tdoa_delay_int()
    test_tdoa_delay_frac_phat()



import unittest

import numpy as np
import pyroomacoustics as pra


def make_filters(n_mics):
    # Location of original source
    azimuth = 61. / 180. * np.pi  # 60 degrees

    # algorithms parameters
    c = 343.
    fs = 16000

    # circular microphone array, 6 mics, radius 15 cm
    R = pra.circular_2D_array([0, 0], n_mics, 0., 0.15)

    # propagation filter bank
    propagation_vector = -np.array([np.cos(azimuth), np.sin(azimuth)])
    delays = np.dot(R.T, propagation_vector) / c * fs  # in fractional samples
    filter_bank = pra.fractional_delay_filter_bank(delays)

    return filter_bank


class TestIssue87(unittest.TestCase):

    def test_12_mics(self):
        # this was working
        make_filters(12)

    def test_6_mics(self):
        # but this failed
        make_filters(6)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
# This file contains code related to acoustics (critical / octave bands, etc)
# Copyright (C) 2019  Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import itertools
import math

import numpy as np
from scipy.fftpack import dct
from scipy.interpolate import interp1d
from scipy.signal import butter, fftconvolve, sosfiltfilt

from .parameters import constants
from .transform import stft


def binning(S, bands):
    """
    This function computes the sum of all columns of S in the subbands
    enumerated in bands
    """
    B = np.zeros((S.shape[0], len(bands)), dtype=S.dtype)
    for i, b in enumerate(bands):
        B[:, i] = np.mean(S[:, b[0] : b[1]], axis=1)

    return B


def bandpass_filterbank(bands, fs=1.0, order=8, output="sos"):
    """
    Create a bank of Butterworth bandpass filters

    Parameters
    ----------
    bands: array_like, shape == (n, 2)
        The list of bands ``[[flo1, fup1], [flo2, fup2], ...]``
    fs: float, optional
        Sampling frequency (default 1.)
    order: int, optional
        The order of the IIR filters (default: 8)
    output: {'ba', 'zpk', 'sos'}
        Type of output: numerator/denominator ('ba'), pole-zero ('zpk'), or
        second-order sections ('sos'). Default is 'ba'.

    Returns
    -------
    b, a : ndarray, ndarray
        Numerator (b) and denominator (a) polynomials of the IIR filter. Only
        returned if output='ba'.
    z, p, k : ndarray, ndarray, float
        Zeros, poles, and system gain of the IIR filter transfer function. Only
        returned if output='zpk'.
    sos : ndarray
        Second-order sections representation of the IIR filter. Only returned
        if output=='sos'.
    """

    filters = []
    nyquist = fs / 2.0

    for band in bands:

        # remove bands above nyquist frequency
        if band[0] >= nyquist:
            raise ValueError("Bands should be below Nyquist frequency")

        # Truncate the highest band to Nyquist frequency
        norm_band = np.minimum(0.99, np.array(band) / nyquist)

        # Compute coefficients
        coeffs = butter(order / 2, norm_band, "bandpass", output=output)
        filters.append(coeffs)

    return filters


def octave_bands(fc=1000, third=False, start=0.0, n=8):
    """
    Create a bank of octave bands

    Parameters
    ----------
    fc : float, optional
        The center frequency
    third : bool, optional
        Use third octave bands (default False)
    start : float, optional
        Starting frequency for octave bands in Hz (default 0.)
    n : int, optional
        Number of frequency bands (default 8)
    """

    div = 1
    if third:
        div = 3

    # Octave Bands
    fcentre = fc * (
        2.0 ** (np.arange(start * div, (start + n) * div - (div - 1)) / div)
    )
    fd = 2 ** (0.5 / div)
    bands = np.array([[f / fd, f * fd] for f in fcentre])

    return bands, fcentre


class OctaveBandsFactory(object):
    """
    A class to process uniformly all properties that are defined on octave
    bands.

    Each property is stored for an octave band.

    Attributes
    ----------
    base_freq: float
        The center frequency of the first octave band
    fs: float
        The target sampling frequency
    n_bands: int
        The number of octave bands needed to cover from base_freq to fs / 2
        (i.e. floor(log2(fs / base_freq)))
    bands: list of tuple
        The list of bin boundaries for the octave bands
    centers
        The list of band centers
    all_materials: list of Material
        The list of all Material objects created by the factory

    Parameters
    ----------
    base_frequency: float, optional
        The center frequency of the first octave band (default: 125 Hz)
    fs: float, optional
        The sampling frequency used (default: 16000 Hz)
    third_octave: bool, optional
        Use third octave bands if True (default: False)
    """

    def __init__(self, base_frequency=125.0, fs=16000, n_fft=512):

        self.base_freq = base_frequency
        self.fs = fs
        self.n_fft = n_fft

        # compute the number of bands
        self.n_bands = math.floor(np.log2(fs / base_frequency))

        self.bands, self.centers = octave_bands(
            fc=self.base_freq, n=self.n_bands, third=False
        )

        self._make_filters()

    def get_bw(self):
        """ Returns the bandwidth of the bands """
        return np.array([b2 - b1 for b1, b2 in self.bands])

    def analysis(self, x, band=None):
        """
        Process a signal x through the filter bank

        Parameters
        ----------
        x: ndarray (n_samples)
            The input signal

        Returns
        -------
        ndarray (n_samples, n_bands)
            The input signal filters through all the bands
        """

        if band is None:
            bands = range(self.filters.shape[1])
        else:
            bands = [band]

        output = np.zeros((x.shape[0], len(bands)), dtype=x.dtype)

        for i, b in enumerate(bands):
            output[:, i] = fftconvolve(x, self.filters[:, b], mode="same")

        if output.shape[1] == 1:
            return output[:, 0]
        else:
            return output

    def __call__(self, coeffs=0.0, center_freqs=None, interp_kind="linear", **kwargs):
        """
        Takes as input a list of values with optional corresponding center frequency.
        Returns a list with the correct number of octave bands. Interpolation and
        extrapolation are used to fill in the missing values.

        Parameters
        ----------
        coeffs: list
            A list of values to use for the octave bands
        center_freqs: list, optional
            The optional list of center frequencies
        interp_kind: str
            Specifies the kind of interpolation as a string (‘linear’,
            ‘nearest’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’,
            ‘next’, where ‘zero’, ‘slinear’, ‘quadratic’ and ‘cubic’ refer to a
            spline interpolation of zeroth, first, second or third order;
            ‘previous’ and ‘next’ simply return the previous or next value of
            the point) or as an integer specifying the order of the spline
            interpolator to use. Default is ‘linear’.
        """

        if not isinstance(coeffs, (list, np.ndarray)):
            # when the parameter is a scalar just do flat extrapolation
            ret = [coeffs] * self.n_bands

        if len(coeffs) == 1:
            ret = coeffs * int(self.n_bands)

        else:
            # by default infer the center freq to be the low ones
            if center_freqs is None:
                center_freqs = self.centers[: len(coeffs)]

            # create the interpolator in log domain
            interpolator = interp1d(
                np.log2(center_freqs),
                coeffs,
                fill_value="extrapolate",
                kind=interp_kind,
            )
            ret = interpolator(np.log2(self.centers))

            # now clip between 0. and 1.
            ret[ret < 0.0] = 0.0
            ret[ret > 1.0] = 1.0

        return ret

    def _make_filters(self):
        """
        Create the band-pass filters for the octave bands

        Parameters
        ----------
        order: int, optional
            The order of the IIR filters (default: 8)
        output: {'ba', 'zpk', 'sos'}
            Type of output: numerator/denominator ('ba'), pole-zero ('zpk'), or
            second-order sections ('sos'). Default is 'ba'.

        Returns
        -------
        A list of callables that will each apply one of the band-pass filters
        """

        """
        filter_bank = bandpass_filterbank(
            self.bands, fs=self.fs, order=order, output=output
        )

        return [lambda sig: sosfiltfilt(bpf, sig) for bpf in filter_bank]
        """

        # This seems to work only for Octave bands out of the box
        centers = self.centers
        n = len(self.centers)

        new_bands = [[centers[0] / 2, centers[1]]]
        for i in range(1, n - 1):
            new_bands.append([centers[i - 1], centers[i + 1]])
        new_bands.append([centers[-2], self.fs / 2])

        n_freq = self.n_fft // 2 + 1
        freq_resp = np.zeros((n_freq, n))
        freq = np.arange(n_freq) / self.n_fft * self.fs

        for b, (band, center) in enumerate(zip(new_bands, centers)):
            lo = np.logical_and(band[0] <= freq, freq < center)
            freq_resp[lo, b] = 0.5 * (1 + np.cos(2 * np.pi * freq[lo] / center))

            if b != n - 1:
                hi = np.logical_and(center <= freq, freq < band[1])
                freq_resp[hi, b] = 0.5 * (1 - np.cos(2 * np.pi * freq[hi] / band[1]))
            else:
                hi = center <= freq
                freq_resp[hi, b] = 1.0

        filters = np.fft.fftshift(
            np.fft.irfft(freq_resp, n=self.n_fft, axis=0), axes=[0],
        )

        # remove the first sample to make them odd-length symmetric filters
        self.filters = filters[1:, :]


def critical_bands():
    """
    Compute the Critical bands as defined in the book:
    Psychoacoustics by Zwicker and Fastl. Table 6.1 p. 159
    """

    # center frequencies
    fc = [
        50,
        150,
        250,
        350,
        450,
        570,
        700,
        840,
        1000,
        1170,
        1370,
        1600,
        1850,
        2150,
        2500,
        2900,
        3400,
        4000,
        4800,
        5800,
        7000,
        8500,
        10500,
        13500,
    ]
    # boundaries of the bands (e.g. the first band is from 0Hz to 100Hz
    # with center 50Hz, fb[0] to fb[1], center fc[0]
    fb = [
        0,
        100,
        200,
        300,
        400,
        510,
        630,
        770,
        920,
        1080,
        1270,
        1480,
        1720,
        2000,
        2320,
        2700,
        3150,
        3700,
        4400,
        5300,
        6400,
        7700,
        9500,
        12000,
        15500,
    ]

    # now just make pairs
    bands = [[fb[j], fb[j + 1]] for j in range(len(fb) - 1)]

    return np.array(bands), fc


def bands_hz2s(bands_hz, Fs, N, transform="dft"):
    """
    Converts bands given in Hertz to samples with respect to a given sampling
    frequency Fs and a transform size N an optional transform type is used to
    handle DCT case.
    """

    # set the bin width
    if transform == "dct":
        B = Fs / 2 / N
    else:
        B = Fs / N

    # upper limit of the frequency range
    limit = min(Fs / 2, bands_hz[-1, 1])

    # Convert from Hertz to samples for all bands
    bands_s = [np.around(band / B) for band in bands_hz if band[0] <= limit]

    # Last band ends at N/2
    bands_s[-1][1] = N / 2

    # remove duplicate, if any, (typically, if N is small and Fs is large)
    j = 0
    while j < len(bands_s) - 1:
        if bands_s[j][0] == bands_s[j + 1][0]:
            bands_s.pop(j)
        else:
            j += 1

    return np.array(bands_s, dtype=np.int)


def melscale(f):
    """ Converts f (in Hertz) to the melscale defined according to Huang-Acero-Hon (2.6) """
    return 1125.0 * np.log(1 + f / 700.0)


def invmelscale(b):
    """ Converts from melscale to frequency in Hertz according to Huang-Acero-Hon (6.143) """
    return 700.0 * (np.exp(b / 1125.0) - 1)


def melfilterbank(M, N, fs=1, fl=0.0, fh=0.5):
    """
    Returns a filter bank of triangular filters spaced according to mel scale

    We follow Huang-Acera-Hon 6.5.2

    Parameters
    ----------
    M : (int)
        The number of filters in the bank
    N : (int)
        The length of the DFT
    fs : (float) optional
        The sampling frequency (default 8000)
    fl : (float)
        Lowest frequency in filter bank as a fraction of fs (default 0.)
    fh : (float)
        Highest frequency in filter bank as a fraction of fs (default 0.5)

    Returns
    -------
    An M times int(N/2)+1 ndarray that contains one filter per row
    """

    # all center frequencies of the filters
    f = (N / fs) * invmelscale(
        melscale(fl * fs)
        + (np.arange(M + 2) * (melscale(fh * fs) - melscale(fl * fs)) / (M + 1))
    )

    # Construct the triangular filter bank
    H = np.zeros((M, N // 2 + 1))
    k = np.arange(N // 2 + 1)
    for m in range(1, M + 1):
        I = np.where(np.logical_and(f[m - 1] < k, k < f[m]))
        H[m - 1, I] = (
            2 * (k[I] - f[m - 1]) / ((f[m + 1] - f[m - 1]) * (f[m] - f[m - 1]))
        )
        I = np.where(np.logical_and(f[m] <= k, k < f[m + 1]))
        H[m - 1, I] = (
            2 * (f[m + 1] - k[I]) / ((f[m + 1] - f[m - 1]) * (f[m + 1] - f[m]))
        )

    return H


def mfcc(x, L=128, hop=64, M=14, fs=8000, fl=0.0, fh=0.5):
    """
    Computes the Mel-Frequency Cepstrum Coefficients (MFCC) according
    to the description by Huang-Acera-Hon 6.5.2 (2001)
    The MFCC are features mimicing the human perception usually
    used for some learning task.

    This function will first split the signal into frames, overlapping
    or not, and then compute the MFCC for each frame.

    Parameters
    ----------
    x : (nd-array)
        Input signal
    L : (int)
        Frame size (default 128)
    hop : (int)
        Number of samples to skip between two frames (default 64)
    M : (int)
        Number of mel-frequency filters (default 14)
    fs : (int)
        Sampling frequency (default 8000)
    fl : (float)
        Lowest frequency in filter bank as a fraction of fs (default 0.)
    fh : (float)
        Highest frequency in filter bank as a fraction of fs (default 0.5)

    Return
    ------
    The MFCC of the input signal
    """

    # perform STFT, X contains frames in rows
    X = stft.analysis(x, L, hop, transform=np.fft.rfft)

    # get and apply the mel filter bank
    # and compute log energy
    H = melfilterbank(M, L, fs=fs, fl=fl, fh=fh)
    S = np.log(np.dot(H, np.abs(X.T) ** 2))

    # Now take DCT of the result
    C = dct(S, type=2, n=M, axis=0)

    return C


def inverse_sabine(rt60, room_dim, c=None):
    """
    Given the desired reverberation time (RT60, i.e. the time for the energy to
    drop by 60 dB), the dimensions of a rectangular room (shoebox), and sound
    speed, computes the energy absorption coefficient and maximum image source
    order needed. The speed of sound used is the package wide default (in
    :py:data:`~pyroomacoustics.parameters.constants`).

    Parameters
    ----------
    rt60: float
        desired RT60 (time it takes to go from full amplitude to 60 db decay) in seconds
    room_dim: list of floats
        list of length 2 or 3 of the room side lengths
    c: float
        speed of sound

    Returns
    -------
    absorption: float
        the energy absorption coefficient to be passed to room constructor
    max_order: int
        the maximum image source order necessary to achieve the desired RT60
    """

    if c is None:
        c = constants.get("c")

    # finding image sources up to a maximum order creates a (possibly 3d) diamond
    # like pile of (reflected) rooms. now we need to find the image source model order
    # so that reflections at a distance of at least up to ``c * rt60`` are included.
    # one possibility is to find the largest sphere (or circle in 2d) that fits in the
    # diamond. this is what we are doing here.
    R = []
    for l1, l2 in itertools.combinations(room_dim, 2):
        R.append(l1 * l2 / np.sqrt(l1 ** 2 + l2 ** 2))

    V = np.prod(room_dim)  # area (2d) or volume (3d)
    # "surface" computation is diff for 2d and 3d
    if len(room_dim) == 2:
        S = 2 * np.sum(room_dim)
        sab_coef = 12  # the sabine's coefficient needs to be adjusted in 2d
    elif len(room_dim) == 3:
        S = 2 * np.sum([l1 * l2 for l1, l2 in itertools.combinations(room_dim, 2)])
        sab_coef = 24

    e_absorption = (
        sab_coef * np.log(10) * V / (c * S * rt60)
    )  # absorption in power (sabine)

    if e_absorption > 1.0:
        raise ValueError(
            "evaluation of parameters failed. room may be too large for required RT60."
        )

    # the int cast is only needed for python 2.7
    # math.ceil returns int for python 3.5+
    max_order = int(math.ceil(c * rt60 / np.min(R) - 1))

    return e_absorption, max_order


def rt60_eyring(S, V, a, m, c):
    """
    This is the Eyring formula for estimation of the reverberation time.

    Parameters
    ----------
    S:
        the total surface of the room walls in m^2
    V:
        the volume of the room in m^3
    a: float
        the equivalent absorption coefficient ``sum(a_w * S_w) / S`` where ``a_w`` and ``S_w`` are the absorption and surface of wall ``w``, respectively.
    m: float
        attenuation constant of air
    c: float
        speed of sound in m/s

    Returns
    -------
    float
        The estimated reverberation time (RT60)
    """

    return -(24 * np.log(10) / c) * V / (S * np.log(1 - a) + 4 * m * V)


def rt60_sabine(S, V, a, m, c):
    """
    This is the Eyring formula for estimation of the reverberation time.

    Parameters
    ----------
    S:
        the total surface of the room walls in m^2
    V:
        the volume of the room in m^3
    a: float
        the equivalent absorption coefficient ``sum(a_w * S_w) / S`` where ``a_w`` and ``S_w`` are the absorption and surface of wall ``w``, respectively.
    m: float
        attenuation constant of air
    c: float
        speed of sound in m/s

    Returns
    -------
    float
        The estimated reverberation time (RT60)
    """

    return (24 * np.log(10) / c) * V / (a * S + 4 * m * V)



from __future__ import division, print_function

import numpy as np
from scipy import signal

from .physics import calculate_speed_of_sound

class DelayCalibration:

    def __init__(self, fs, pad_time=0., mls_bits=16, repeat=1, 
                 temperature=25., humidity=50., pressure=1000.):
        '''
        Initialize the delay calibration object

        Parameters
        ----------
        fs : int
            Sampling frequency to use
        pad_time : float
            Duration of silence to add at the end of the test signal (i.e. expected T60)
        mls_bits : int
            Number of bits to use for the maximum length sequence
        repeat : int
            Number of repetition of the measurement before averaging
        temperature : float
            Room temperature
        humidity : float
            Ambient humidity
        pressure : float
            Atmospheric pressure
        '''
            

        self.fs = fs
        self.mls_bits = mls_bits
        self.repeat = repeat
        self.pad_time = pad_time

        self.temperature = temperature
        self.humidity = humidity
        self.pressure = pressure
        self.c = calculate_speed_of_sound(temperature, humidity, pressure)

    def run(self, distance=0., ch_in=None, ch_out=None, oversampling=1):
        '''
        Run the calibration. Plays a maximum length sequence and cross correlate
        the signals to find the time delay.

        Parameters
        ----------
        distance : float, optional
            Distance between the speaker and microphone
        ch_in : int, optional
            The input channel to use. If not specified, all channels are calibrated
        ch_out : int, optional
            The output channel to use. If not specified, all channels are calibrated
        '''

        if ch_out is None:
            ch_out = [0]

        # create the maximum length sequence
        mls = 0.95*np.array(2*signal.max_len_seq(self.mls_bits)[0] - 1, dtype=np.float32)

        # output signal
        s = np.zeros((mls.shape[0] + int(self.pad_time*self.fs), sd.default.channels[1]), dtype=np.float32)

        # placeholder for delays
        delays = np.zeros((sd.default.channels[0], len(ch_out)))

        try:
            import matplotlib.pyplot as plt
        except ImportError:
            import warnings
            warnings.warn('Matplotlib is required for plotting')
            return

        for och in ch_out:
            # play and record the signal simultaneously
            s[:mls.shape[0], och] = 0.1 * mls
            rec_sig = sd.playrec(s, self.fs, channels=sd.default.channels[0], blocking=True)
            s[:,och] = 0

            for ich in range(rec_sig.shape[1]):
                xcorr = signal.correlate(rec_sig[:,ich], mls, mode='valid')
                delays[ich,och] = np.argmax(xcorr)
                plt.plot(xcorr)
                plt.show()

        # subtract distance
        delays -= int(distance / self.c * self.fs)

        return delays


if __name__ == "__main__":

    try:
        import sounddevice as sd

        sd.default.device = (2,2)
        sd.default.channels = (1,2)
        dc = DelayCalibration(48000, mls_bits=16, pad_time=0.5, repeat=1, temperature=25.6, humidity=30.)

        delays = dc.run(ch_out=[0,1])

        print(delays)
    except:
        raise ImportError('Sounddevice package must be installed to run that script.')




# Test of Wall reflection side test
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import unittest
import numpy as np
import pyroomacoustics as pra

eps = 1e-6

corners = {
        '3d' : np.array([
            [-1, 1, 1],
            [1, -1, 1],
            [1, 1, -1],
            ]),
        '2d' : np.array([
            [ -1, 1 ],
            [ -2, 2 ],
            ]),
        }

points = {
        '3d_up' : {
            'd' : '3d',
            'p' : [-1,-1,-1],
            'expect' : -1,
            'reflect' : [5/3, 5/3, 5/3],
            },
        '3d_down' : {
            'd' : '3d',
            'p' : [1,1,1],
            'expect' : 1,
            'reflect' : [-1/3, -1/3, -1/3],
            },
        '3d_on' : {
            'd' : '3d',
            'p' : [1/3, 1/3, 1/3],
            'expect' : 0,
            'reflect' : [1/3, 1/3, 1/3],
            },

        '2d_down' : {
            'd' : '2d',
            'p' : [-2, 1],
            'expect' : -1,
            'reflect' : [2, -1],
            },
        '2d_up' : {
            'd' : '2d',
            'p' : [2, -1],
            'expect' : 1,
            'reflect' : [-2, 1],
            },
        '2d_on' : {
            'd' : '2d',
            'p' : [0.5, 1],
            'expect' : 0,
            'reflect' : [0.5, 1],
            },
        }


def run_side(label):
    p = points[label]['p']
    r_exp = points[label]['expect']
    d = points[label]['d']

    wall = pra.wall_factory(corners[d], [0.1], [0.1])
    r = wall.side(p)

    print('{}: returned={} expected={}'.format(label, r, r_exp))
    return r == r_exp


def run_reflect(label):

    p = points[label]['p']
    p_refl = np.array(points[label]['reflect'])
    r_exp = points[label]['expect']
    d = points[label]['d']

    wall = pra.wall_factory(corners[d], [0.1], [0.1])

    x = np.zeros(wall.dim, dtype=np.float32)
    wall.reflect(p, x)

    err = np.linalg.norm(x - p_refl) < eps

    print('{}: error={}'.format(label, err))
    return err


class TestUtilityRoutines(unittest.TestCase):

    def test_side_3d_up(self):
        ret = run_side('3d_up')
        self.assertTrue(ret)

    def test_side_3d_down(self):
        ret = run_side('3d_down')
        self.assertTrue(ret)

    def test_side_3d_on(self):
        ret = run_side('3d_on')
        self.assertTrue(ret)

    def test_side_2d_up(self):
        ret = run_side('2d_up')
        self.assertTrue(ret)

    def test_side_2d_down(self):
        ret = run_side('2d_down')
        self.assertTrue(ret)

    def test_side_2d_on(self):
        ret = run_side('2d_on')
        self.assertTrue(ret)

    def test_reflect_3d_up(self):
        ret = run_reflect('3d_up')
        self.assertTrue(ret)

    def test_reflect_3d_down(self):
        ret = run_reflect('3d_down')
        self.assertTrue(ret)

    def test_reflect_3d_on(self):
        ret = run_reflect('3d_on')
        self.assertTrue(ret)

    def test_reflect_2d_up(self):
        ret = run_reflect('2d_up')
        self.assertTrue(ret)

    def test_reflect_2d_down(self):
        ret = run_reflect('2d_down')
        self.assertTrue(ret)

    def test_reflect_2d_on(self):
        ret = run_reflect('2d_on')
        self.assertTrue(ret)

    def test_reflected_end2D(self):

        eps = 0.001
        start = [1,3]

        hit = [5,3]
        #normal = [-1, -1]
        corners = np.array([
            [ 6, 4 ],
            [ 4, 6 ],
            ])
        wall = pra.wall_factory(corners, [0.1], [0.1])
        length = 4

        res = wall.normal_reflect(start, hit, length)
        self.assertTrue(np.allclose(res, [5.,-1.], atol=eps))

    def test_reflected_end3D(self):

        eps = 0.001
        start = [1,1,1]
        hit = [-1, 1, 3]
        #normal = [1, 0, 0]
        corners = np.array([
            [ -1, -1, -1, -1, ],
            [  0,  2,  2,  0, ],
            [  2,  2,  4,  4, ],
            ])
        wall = pra.wall_factory(corners, [0.1], [0.1])

        length = 2*np.sqrt(2)

        res = wall.normal_reflect(start, hit, length)
        self.assertTrue(np.allclose(res, [1,1,5], atol=eps))


if __name__ == '__main__':
    unittest.main()



import pyroomacoustics as pra
from pyroomacoustics.parameters import _calculate_temperature


def test_set_temperature_wo_humidity():
    room = pra.ShoeBox(p=[3, 3, 3], temperature=21)
    assert room.physics.H == 0.0


def test_set_humidity_wo_temperature():
    room = pra.ShoeBox(p=[3, 3, 3], humidity=21)
    assert room.physics.T == _calculate_temperature(
        room.physics.get_sound_speed(), room.physics.H
    )


if __name__ == '__main__':
    test_set_temperature_wo_humidity()
    test_set_humidity_wo_temperature()




import numpy as np
import pyroomacoustics as pra

def test_room_volume():

    eps = 0.00001


    # Create the 2D L-shaped room from the floor polygon
    pol = 4 * np.array([[0, 0], [0, 1], [2, 1], [2, 0.5], [1, 0.5], [1, 0]]).T
    r_absor = 0.1
    room = pra.Room.from_corners(pol, fs=16000, max_order=6, absorption=r_absor)

    # Create the 3D room by extruding the 2D by 3 meters
    room.extrude(3., absorption=r_absor)

    assert np.allclose(room.get_volume(), 72, atol=eps)


def test_walls_area():
    eps = 0.00001

    # Create the 2D L-shaped room from the floor polygon
    pol = 4 * np.array([[0, 0], [0, 1], [2, 1], [2, 0.5], [1, 0.5], [1, 0]]).T
    r_absor = 0.1
    room = pra.Room.from_corners(pol, fs=16000, max_order=6, absorption=r_absor)

    # Create the 3D room by extruding the 2D by 3 meters
    room.extrude(3., absorption=r_absor)

    assert np.allclose(room.wall_area(room.walls[0]), 6, atol=eps)
    assert np.allclose(room.wall_area(room.walls[1]), 12, atol=eps)
    assert np.allclose(room.wall_area(room.walls[2]), 6, atol=eps)
    assert np.allclose(room.wall_area(room.walls[3]), 24, atol=eps)
    assert np.allclose(room.wall_area(room.walls[4]), 12, atol=eps)
    assert np.allclose(room.wall_area(room.walls[5]), 12, atol=eps)
    assert np.allclose(room.wall_area(room.walls[6]), 24, atol=eps)
    assert np.allclose(room.wall_area(room.walls[7]), 24, atol=eps)

if __name__ == '__main__':
    test_room_volume()
    test_walls_area()



import numpy as np
import pyroomacoustics as pra


def test_issue_115_ism_breaking():
    """
    When a source was too close to the microphone, the time-of-flight
    might be smaller than the delay due to the fractionnal delay filter
    used to create the impulse response.
    It is then necessary to add this delay to the rir filter to ensure
    no runtime error.
    """
    print("Test with source close to microphone.")
    shoebox = pra.ShoeBox(
        [9.29447785567344, 6.529510207957697, 4.4677460263160995],
        materials=pra.Material(energy_absorption=0.1675976883006225),
        fs=16000,
        max_order=17
    )
    source_loc = [5.167674641605016, 4.379726875714017, 2.9190423403507504]
    shoebox.add_source(source_loc)
    noise_loc = [8.47420884677372, 5.675261722911696, 1.2040578622058364]
    shoebox.add_source(noise_loc)
    R = np.array([[8.571318246865648],
                  [5.799718630723678],
                  [1.3702254938278977]])
    print('mic - source distance : {} m'.format(
        np.sqrt(sum((np.array(source_loc) - np.squeeze(R))**2)))
    )
    print('mic - noise distance : {} m'.format(
        np.sqrt(sum((np.array(noise_loc) - np.squeeze(R))**2)))
    )
    shoebox.add_microphone_array(
        pra.MicrophoneArray(R, shoebox.fs))
    shoebox.compute_rir()


def test_issue_115_rt_breaking():
    """
    As background, only ray tracing only starts to be active for rays that run
    longer than the maximum ISM order.
    The problem happen when the ISM order is very high (here 17),
    then, in some circumstances, it is possible that no ray travels longer
    than that. Then the histogram is empty and an error happen.
    """
    print("Test with high order ISM")
    shoebox = pra.ShoeBox(
        [4.232053263716528, 3.9244954007318853, 5.563810437305445],
        materials=pra.Material(energy_absorption=0.6965517438548237),
        fs=16000,
        max_order=17,
        ray_tracing=True,
    )
    source_loc = [1.2028020579854695, 2.2980760894630676, 2.0654520390433984]
    shoebox.add_source(source_loc)
    R = np.array([[1.8062807887952617], [2.7793113278109454], [1.42966428606882]])
    print(
        "mic - source distance : {} m".format(
            np.sqrt(sum((np.array(source_loc) - np.squeeze(R)) ** 2))
        )
    )
    shoebox.add_microphone_array(pra.MicrophoneArray(R, shoebox.fs))
    shoebox.compute_rir()


if __name__ == "__main__":
    test_issue_115_rt_breaking()
    test_issue_115_ism_breaking()



from __future__ import division, print_function

import numpy as np
from scipy import linalg as la
from .point_cloud import PointCloud

try:
    import mklfft as fft
except ImportError:
    import numpy.fft as fft

def tdoa_loc(R, tdoa, c, x0=None):
    '''
    TDOA based localization

    Parameters
    ----------
    R : ndarray
        A 3xN array of 3D points
    tdoa : ndarray
        A length N array of tdoa
    c : float
        The speed of sound

    Reference
    ---------
    Steven Li, TDOA localization
    '''
    tau = tdoa - tdoa[0]

    # eliminate 0 tdoa
    I = tau != 0.
    I[0] = True  # keep mic 0! (reference, tdoa always 0)
    tau = tau[I]
    R = R[:,I]

    # Need two ref points
    r0 = R[:,0:1]
    r1 = R[:,1:2]
    rm = R[:,2:]

    n0 = la.norm(r0)**2
    n1 = la.norm(r1)**2
    nm = la.norm(rm, axis=0)**2

    # Build system matrices
    # Steven Li's equations
    ABC = 2 * ( rm - r0 ) / (c * tau[2:]) - 2 * (r1 - r0) / (c * tau[1])
    D = c * tau[1] - c * tau[2:] + (nm - n0) / (c * tau[2:]) - (n1 - n0) / (c * tau[1])

    loc = la.lstsq(ABC.T, D)[0]
    '''

    from scipy.optimize import leastsq

    def f(r, *args):

        R = args[0]
        c = args[1]
        tdoa = args[2]

        res = la.norm(R - r[:3,None], axis=0) - (r[3] + c * tau)

        return res

    def Jf(r, *args):

        R = args[0]
        c = args[1]
        tdoa = args[2]

        delta = r[:3,None] - R
        norm = la.norm(delta, axis=0)

        J = np.zeros((R.shape[0]+1, R.shape[1]))
        J[:3,:] = (delta / norm)
        J[3,:] = -1.

        return J

    init = f(x0, R[:,1:], c, tdoa[1:])
    sol = leastsq(f, x0, args=(R[:,1:],c,tdoa[1:]), Dfun=Jf, full_output=True, maxfev=10000, col_deriv=True)
    print sol[2]['nfev']
    print sol[1]
    print np.sum(f(sol[0], R[:,1:], c, tdoa[1:])**2) / np.sum(init**2)

    loc = sol[0][:3]
    print 'distance offset',sol[0][3]
    '''

    return loc

def tdoa(x1, x2, interp=1, fs=1, phat=True):
    '''
    This function computes the time difference of arrival (TDOA)
    of the signal at the two microphones. This in turns is used to infer
    the direction of arrival (DOA) of the signal.
    
    Specifically if s(k) is the signal at the reference microphone and
    s_2(k) at the second microphone, then for signal arriving with DOA
    theta we have
    
    s_2(k) = s(k - tau)
    
    with
    
    tau = fs*d*sin(theta)/c
    
    where d is the distance between the two microphones and c the speed of sound.
    
    We recover tau using the Generalized Cross Correlation - Phase Transform (GCC-PHAT)
    method. The reference is
    
    Knapp, C., & Carter, G. C. (1976). The generalized correlation method for estimation of time delay. 
    
    Parameters
    ----------
    x1 : nd-array
        The signal of the reference microphone
    x2 : nd-array
        The signal of the second microphone
    interp : int, optional (default 1)
        The interpolation value for the cross-correlation, it can
        improve the time resolution (and hence DOA resolution)
    fs : int, optional (default 44100 Hz)
        The sampling frequency of the input signal
        
    Return
    ------
    theta : float
        the angle of arrival (in radian (I think))
    pwr : float
        the magnitude of the maximum cross correlation coefficient
    delay : float
        the delay between the two microphones (in seconds)
    '''

    # zero padded length for the FFT
    n = (x1.shape[0]+x2.shape[0]-1)
    if n % 2 != 0:
        n += 1

    # Generalized Cross Correlation Phase Transform
    # Used to find the delay between the two microphones
    # up to line 71
    X1 = fft.rfft(np.array(x1, dtype=np.float32), n=n)
    X2 = fft.rfft(np.array(x2, dtype=np.float32), n=n)

    if phat:
        X1 /= np.abs(X1)
        X2 /= np.abs(X2)

    cc = fft.irfft(X1*np.conj(X2), n=interp*n)

    # maximum possible delay given distance between microphones
    t_max = n // 2 + 1

    # reorder the cross-correlation coefficients
    cc = np.concatenate((cc[-t_max:],cc[:t_max]))

    # pick max cross correlation index as delay
    tau = np.argmax(np.abs(cc))
    pwr = np.abs(cc[tau])
    tau -= t_max  # because zero time is at the center of the array

    return tau / (fs*interp)


def edm_line_search(R, tdoa, bounds, steps):
    '''
    We have a number of points of know locations and have the TDOA measurements
    from an unknown location to the known point.
    We perform an EDM line search to find the unknown offset to turn TDOA to TOA.

    Parameters
    ----------
    R : ndarray
        An ndarray of 3xN where each column is the location of a point
    tdoa : ndarray
        A length N vector containing the tdoa measurements from uknown location to known ones
    bounds : ndarray
        Bounds for the line search
    step : float
        Step size for the line search
    '''

    dim = R.shape[0]

    pc = PointCloud(X=R)

    # use point 0 as reference
    dif = tdoa - tdoa.min()

    # initialize EDM
    D = np.zeros((pc.m+1, pc.m+1))
    D[:-1,:-1] = pc.EDM()

    # distance offset to search
    d = np.linspace(bounds[0], bounds[1], steps)

    # sum of eigenvalues that should be zero
    #cost = np.zeros((d.shape[0], D.shape[0]))
    cost = np.zeros(*d.shape)

    for i in range(d.shape[0]):
        D[-1,:-1] = D[:-1,-1] = (dif + d[i])**2
        w = np.sort(np.abs(la.eigh(D, eigvals_only=True)))
        #w = la.eigh(D, eigvals_only=True, eigvals=(D.shape[0]-6,D.shape[0]-6))
        cost[i] = np.sum(w[:D.shape[0]-5])

    return cost, d



# Test of polygon point inclusion test
# Copyright (C) 2019  Robin Scheibler, Cyril Cadoux, Sidney Barthe
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
from __future__ import division

import numpy as np
import pyroomacoustics as pra

polygons = [
        np.array([  # this one is clockwise
            [0, 4, 4, 0,],
            [0, 0, 4, 4,],
            ]),

        np.array([  # this one is clockwise!
            [0, 0, 1, 1, 3, 3],
            [0, 1, 1, 2, 2, 0],
            ]),

        np.array([  # this one is clockwise!
            [0, 1, 1, 3, 3],
            [0, 1, 2, 2, 0],
            ]),
        ]

cases = {
        'inside' : {
            'pol' : 0,
            'p' : [2,2],
            'ret' : 0,
            },

        'on_border' : {
            'pol' : 0,
            'p' : [0,2],
            'ret' : 1,
            },

        'on_corner' : {
            'pol' : 0,
            'p' : [4,4],
            'ret' : 1,
            },

        'outside' : {
            'pol' : 0,
            'p' : [5,5],
            'ret' : -1,
            },

        # horizontal wall aligned with point
        'horiz_wall_align' : {
            'pol' : 1,
            'p' : [2,1],
            'ret' : 0,
            },
        # ray is going through vertex
        'ray_through_vertex' : {
            'pol' : 2,
            'p' : [2,1],
            'ret' : 0,
            },

        # point is at the same height as top of polygon, but outside
        'top_outside' : {
            'pol' : 2,
            'p' : [4,2],
            'ret' : -1,
            },
        }

def run_inside_pol(lbl):
    pol = polygons[cases[lbl]['pol']]
    p = cases[lbl]['p']
    r_exp = cases[lbl]['ret']

    ret = pra.libroom.is_inside_2d_polygon(p, pol)

    assert ret == r_exp, '{} : returned={} expected={}'.format(lbl, ret, r_exp)

def test_inside():
    run_inside_pol('inside')

def test_on_border():
    run_inside_pol('on_border')

def test_on_corner():
    run_inside_pol('on_corner')

def test_outside():
    run_inside_pol('outside')

def test_horiz_wall_align():
    run_inside_pol('horiz_wall_align')

def test_ray_through_vertex():
    run_inside_pol('ray_through_vertex')

def test_top_outside():
    run_inside_pol('top_outside')

if __name__ == '__main__':

    test_inside()
    test_on_border()
    test_on_corner()
    test_outside()
    test_top_outside()
    test_horiz_wall_align()
    test_ray_through_vertex()



# Single Channel Noise Removal using the Subspace Approach
# Copyright (C) 2019  Eric Bezzam, Mathieu Lecoq, Gimena Segrelles
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.


import numpy as np


class Subspace(object):
    """
    A class for performing **single channel** noise reduction in the time domain
    via the subspace approach. This implementation is based off of the approach
    presented in:

        Y. Ephraim and H. L. Van Trees, *A signal subspace approach for speech enhancement,*
        IEEE Transactions on Speech and Audio Processing, vol. 3, no. 4, pp. 251-266, Jul 1995.

    Moreover, an adaptation of the subspace approach is implemented here, as
    presented in:

        Y. Hu and P. C. Loizou, *A subspace approach for enhancing speech corrupted by colored noise,*
        IEEE Signal Processing Letters, vol. 9, no. 7, pp. 204-206, Jul 2002.

    Namely, an eigendecomposition is performed on the matrix:

    .. math::

        \Sigma = R_n^{-1} R_y - I,

    where :math:`R_n` is the noise covariance matrix, :math:`R_y` is the
    covariance matrix of the input noisy signal, and :math:`I` is the identity
    matrix. The covariance matrices are estimated from past samples; the number
    of past samples/frames used for estimation can be set with the parameters
    `lookback` and `skip`. A simple energy threshold (`thresh` parameter) is
    used to identify noisy frames.

    The eigenvectors corresponding to the positive eigenvalues of
    :math:`\Sigma` are used to create a linear operator :math:`H_{opt}` in order
    to enhance the noisy input signal :math:`\mathbf{y}`:

    .. math::

        \mathbf{\hat{x}} = H_{opt} \cdot \mathbf{y}.

    The length of :math:`\mathbf{y}` is specified by the parameter `frame_len`;
    :math:`50\%` overlap and add with a Hanning window is used to reconstruct
    the output. A great summary of the approach can be found in the paper by
    Y. Hu and P. C. Loizou under Section III, B.

    Adjusting the factor :math:`\\mu` presents a trade-off between noise
    suppression and distortion.

    Below is an example of how to use this class to emulate a streaming/online
    input. A full example can be found `here <https://github.com/LCAV/pyroomacoustics/blob/master/examples/noise_reduction_subspace.py>`_.
    Depending on your choice for `frame_len`, `lookback`, and `skip`, the
    approach may not be suitable for real-time processing.

    ::

        # initialize Subspace object
        scnr = Subspace(frame_len=256, mu=10, lookback=10, skip=2, thresh=0.01)

        # apply block-by-block
        for n in range(num_blocks):
            denoised_signal = scnr.apply(noisy_input)


    There also exists a "one-shot" function.

    ::

        # import or create `noisy_signal`
        denoised_signal = apply_subspace(noisy_signal, frame_len=256, mu=10,
                                         lookback=10, skip=2, thresh=0.01)


    Parameters
    ----------
    frame_len : int
        Frame length in samples. Note that large values (above 256) will make
        the eigendecompositions very expensive.
    mu : float
        Enhancement factor, larger values suppress more noise but could lead
        to more distortion.
    lookback : int
        How many frames to look back for covariance matrix estimation.
    skip : int
        How many samples to skip when estimating the covariance matrices with
        past samples. `skip=1` will use all possible frames in the estimation.
    thresh : float
        Threshold to distinguish between (signal+noise) and (noise) frames. A
        high value will classify more frames as noise but might remove desired
        signal!
    data_type : 'float32' or 'float64'
        Data type to use in the enhancement procedure. Default is 'float32'.
    """


    def __init__(self, frame_len=256, mu=10, lookback=10, skip=2, thresh=0.01,
                 data_type='float32'):

        if frame_len % 2:
            raise ValueError("Frame length should be even as this method "
                             "performs 50% overlap.")

        if data_type is 'float64':
            data_type = np.float64
        else:
            data_type = np.float32

        self.frame_len = frame_len
        self.mu = mu
        self.dtype = data_type

        self.hop = frame_len // 2
        self.prev_samples = np.zeros(self.hop, dtype=data_type)
        self.prev_output = np.zeros(self.hop, dtype=data_type)
        self.current_out = np.zeros(self.hop, dtype=data_type)
        self.win = np.hanning(frame_len).astype(data_type)

        # enhancement filter parameter
        self.h_opt = np.zeros((frame_len, frame_len), dtype=data_type)

        # estimate (signal+noise) and noise covariance matrices
        self.thresh = thresh
        self.n_samples = self.hop * lookback + frame_len
        self.input_samples = np.zeros(self.n_samples, dtype=data_type)
        self.skip = skip
        self.n_frames = lookback * (self.hop//skip)
        self.n_noise_frames = np.ones(lookback) * (self.hop // skip)
        self.cov_sn = np.zeros((frame_len, frame_len), dtype=data_type)
        self._cov_sn = np.zeros((lookback, frame_len, frame_len),
                                dtype=data_type)
        self.cov_n = np.zeros((frame_len, frame_len), dtype=data_type)
        self._cov_n = np.zeros((lookback, frame_len, frame_len),
                               dtype=data_type)

    def apply(self, new_samples):
        """
        Parameters
        ----------
        new_samples: numpy array
            New array of samples of length `self.hop` in the time domain.

        Returns
        -------
        numpy array
            Denoised samples.
        """

        if len(new_samples) != self.hop:
            raise ValueError("Expected {} samples, got {}."
                             .format(self.hop, len(new_samples)))
        new_samples = new_samples.astype(self.dtype)

        # form input frame, 50% overlap
        input_frame = np.r_[self.prev_samples, new_samples]

        # update covariance matrix estimates
        self.update_cov_matrices(new_samples)

        # compute filter to project to signal subspace
        self.compute_signal_projection()

        # compute output
        denoised_out = self.win * np.dot(self.h_opt, input_frame)

        # update
        self.prev_samples[:] = new_samples
        self.current_out[:] = self.prev_output + denoised_out[:self.hop]
        self.prev_output[:] = denoised_out[self.hop:]

        return self.current_out

    def compute_signal_projection(self):

        sigma = np.linalg.lstsq(self.cov_n, self.cov_sn, rcond=None)[0] \
                - np.eye(self.frame_len)
        eigenvals, eigenvecs = np.linalg.eig(sigma)

        n_pos = sum(eigenvals > 0)
        order = np.argsort(eigenvals, axis=-1)[::-1]
        pos_eigenvals = np.real(eigenvals[order][:n_pos])
        q1 = np.zeros((self.frame_len, self.frame_len))
        for w in range(0, n_pos):
            q1[w, w] = pos_eigenvals[w] / (pos_eigenvals[w] + self.mu)

        v_t = np.transpose(-eigenvecs[:, order])
        self.h_opt[:] = np.real(np.dot(np.dot(np.linalg.pinv(v_t), q1), v_t))
        # self.h_opt = np.dot(np.linalg.lstsq(v_t, q1, rcond=None)[0], v_t)

    def update_cov_matrices(self, new_samples):

        # remove cov of old samples
        self.cov_sn *= self.n_frames
        self.cov_sn -= self._cov_sn[0]

        old_cov_n = self.cov_n.copy()
        self.cov_n *= sum(self.n_noise_frames)
        self.cov_n -= self._cov_n[0]

        # update samples
        self.input_samples = np.roll(self.input_samples, -self.hop)
        self.input_samples[-self.hop:] = new_samples

        # update cov matrices
        self._cov_sn = np.roll(self._cov_sn,  -1, axis=0)
        self._cov_sn[-1, :, :] = np.zeros((self.frame_len, self.frame_len))
        self._cov_n = np.roll(self._cov_n, -1, axis=0)
        self._cov_n[-1, :, :] = np.zeros((self.frame_len, self.frame_len))
        self.n_noise_frames = np.roll(self.n_noise_frames, -1)
        self.n_noise_frames[-1] = 0

        for i in range(0, self.hop, self.skip):
            a = self.n_samples - self.hop - self.frame_len + i
            b = a + self.frame_len
            _noisy_signal = self.input_samples[a:b]
            new_cov = np.outer(_noisy_signal, _noisy_signal).astype(self.dtype)

            # (signal+noise) cov
            self._cov_sn[-1] += new_cov

            # noise cov
            energy = np.std(_noisy_signal) ** 2
            if energy < self.thresh:
                self._cov_n[-1] += new_cov
                self.n_noise_frames[-1] += 1

        # if no new noise frames, use previous
        if self.n_noise_frames[-1] == 0:
            self._cov_n[-1] = old_cov_n
            self.n_noise_frames[-1] = 1

        # compute average for new estimate
        self.cov_sn = (self.cov_sn + self._cov_sn[-1]) / self.n_frames
        self.cov_n = (self.cov_n + self._cov_n[-1]) / sum(self.n_noise_frames)


def apply_subspace(noisy_signal, frame_len=256, mu=10, lookback=10, skip=2,
                   thresh=0.01, data_type=np.float32):
    """
    One-shot function to apply subspace denoising approach.

    Parameters
    ----------
    noisy_signal : numpy array
        Real signal in time domain.
    frame_len : int
        Frame length in samples. Note that large values (above 256) will make
        the eigendecompositions very expensive. 50% overlap is used with
        hanning window.
    mu : float
        Enhancement factor, larger values suppress more noise but could lead
        to more distortion.
    lookback : int
        How many frames to look back for covariance matrix estimation.
    skip : int
        How many samples to skip when estimating the covariance matrices with
        past samples. `skip=1` will use all possible frames in the estimation.
    thresh : float
        Threshold to distinguish between (signal+noise) and (noise) frames. A
        high value will classify more frames as noise and might even remove
        desired signal!
    data_type : 'float32' or 'float64'
        Data type to use in the enhancement procedure. Default is 'float32'.

    Returns
    -------
    numpy array
        Enhanced/denoised signal.
    """

    scnr = Subspace(frame_len, mu, lookback, skip, thresh, data_type)
    processed_audio = np.zeros(noisy_signal.shape)
    n = 0
    hop = frame_len // 2
    while noisy_signal.shape[0] - n >= hop:

        processed_audio[n:n + hop, ] = scnr.apply(noisy_signal[n:n + hop])

        # update step
        n += hop

    return processed_audio





# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

from unittest import TestCase
import numpy as np

import pyroomacoustics as pra

fs = 8000
t0 = 0.
absorption = 0.90
max_order_sim = 2
sigma2_n = 5e-7

corners = np.array([
    [0,5,5,3,3,2,2,0],
    [0,0,5,5,2,2,5,5],
    ])
room = pra.Room.from_corners(
    corners,
    absorption=absorption,
    fs=fs,
    t0=t0,
    max_order=max_order_sim,
    sigma2_awgn=sigma2_n)
    
room.add_source([1, 4], None, 0)

# place 3 microphones in the room
mics = pra.MicrophoneArray(np.array([[4, 2.5, 2.5,],
                                     [4, 1,   4   ]]), fs)
room.add_microphone_array(mics)

# run the image source model
room.image_source_model()

# we order the sources lexicographically
ordering = np.lexsort(room.sources[0].images)
images_found = room.sources[0].images[:,ordering]
visibilities = room.visibility[0][:,ordering]

# This should be all visible sources, from the 3 mic locations
sources = np.array([[ 1., -1.,  1., -3., -1., -1.,  1.],
                    [-6., -4., -4.,  4.,  4.,  6.,  8.]])

# These are the visibilities for each individual microphone
visible_opposite_side = [0, 0, 1, 0, 0, 0, 0]
visible_middle = [1, 1, 1, 1, 1, 1, 1]
visible_outside = [0, 0, 0, 0, 0, 0, 0]

def test_sources():
    assert np.allclose(images_found, sources)

def test_visibility_oppositeSide():
    assert all(visibilities[0] == visible_opposite_side)

def test_visibility_middle():
    assert all(visibilities[1] == visible_middle)

def test_visibility_outside():
    assert all(visibilities[2] == visible_outside)

if __name__ == '__main__':
    test_sources()
    test_visibility_oppositeSide()
    test_visibility_middle()
    test_visibility_outside()




from __future__ import division, print_function

import numpy as np

class Buffer:
    '''
    A simple buffer class with amortized cost

    Parameters
    ----------
    length: int
        buffer length
    dtype: numpy.type
        data type
    '''

    def __init__(self, length=20, dtype=np.float64):

        self.buf = np.zeros(length, dtype=dtype)
        self.len = length
        self.head = self.len

    def push(self, val):
        ''' Add one element at the front of the buffer '''

        # Increase size if the buffer is too small
        if self.head == 0:
            self.buf = np.concatenate((np.zeros(self.len, dtype=self.buf.dtype), self.buf))
            self.head += self.len
            self.len *= 2

        # store value at head
        self.buf[self.head-1] = val

        # move head to next free spot
        self.head -= 1

    def top(self, n):
        ''' Returns the n elements at the front of the buffer from newest to oldest '''

        return self.buf[self.head:self.head+n]

    def flush(self, n):
        ''' Removes the n oldest elements in the buffer '''

        if n > self.len - self.head:
            n = self.len - self.head

        new_head = self.head + n

        # copy the remaining items to the right
        self.buf[new_head:] = self.buf[self.head:-n]

        # move head
        self.head = new_head

    def size(self):
        ''' Returns the number of elements in the buffer '''
        return self.len - self.head

    def __getitem__(self, r):
        ''' Allows to retrieve element at a specific position '''

        # create a view that starts at head
        ptr = self.buf[self.head:]

        # returned desired range
        return ptr[r]

    def __repr__(self):

        if self.head == self.len:
            return '[]'
        else:
            return str(self.buf[self.head:])

class Powers:
    '''
    This class allows to store all powers of a small number
    and get them 'a la numpy' with the bracket operator.
    There is automatic increase when new values are requested

    Parameters
    ----------
    a: float
        the number
    length: int
        the number of integer powers
    dtype: numpy.type, optional
        the data type (typically np.float32 or np.float64)

    Example
    -------

        >>> an = Powers(0.5)
        >>> print(an[4])
        0.0625

    '''

    def __init__(self, a, length=20, dtype=np.float64):

        self.a = dtype(a)
        self.pwr = self.a ** np.arange(length)

    def __getitem__(self, r):

        # find maximum power requested
        if isinstance(r, int):
            high = r + 1
        elif isinstance(r, slice):
            high = r.stop
        elif isinstance(r, list):
            high = max(r) + 1
        else:
            high = int(r + 1)

        # Compute it if needed
        if high > self.pwr.shape[0]:
            self.pwr = np.concatenate((self.pwr, self.a**np.arange(self.pwr.shape[0], high)))

        return self.pwr[r]

    def __repr__(self):
        return str(self.pwr)


class CoinFlipper:
    '''
    This class efficiently generates large number of coin flips.
    Because each call to ``numpy.random.rand`` is a little bit costly,
    it is more efficient to generate many values at once.
    This class does this and stores them in advance. It generates
    new fresh numbers when needed.

    Parameters
    ----------
    p: float, 0 < p < 1
        probability to output a 1
    length: int
        the number of flips to precompute
    '''

    def __init__(self, p, length=10000):

        self.p = p
        self.length = length
        self.buffer = np.random.random(length) < p
        self.dirty_coins = 0

    def fresh_flips(self, n):
        ''' Generates n binary random values now '''

        return np.random.random(n) < self.p

    def flip_all(self):
        ''' Regenerates all the used up values '''

        remaining = self.length - self.dirty_coins
        self.buffer[:self.dirty_coins] = self.fresh_flips(self.dirty_coins)
        self.dirty_coins = 0


    def flip(self, n):
        ''' Get n random binary values from the buffer '''


        # If more flips than computed are requested
        # increase buffer size and flip again
        if n > self.length:
            self.buffer = np.pad(self.buffer, (0, 2 * n - self.length), mode='constant')
            self.buffer[self.length:] = self.fresh_flips(2 * n - self.length)
            self.length = 2 * n

        remaining = self.length - self.dirty_coins
        if remaining < n:
            self.flip_all()

        flips = self.buffer[self.dirty_coins:self.dirty_coins+n]
        self.dirty_coins += n

        return flips




# cython: infer_types=True

import numpy as np
cimport cython

from libc.math cimport floor, ceil

@cython.boundscheck(False)
@cython.wraparound(False)
def fast_rir_builder(
        double [:] rir,
        double [:] time,
        double [:] alpha,
        int [:] visibility,
        int fs,
        int fdl,
        int lut_gran=20,
        ):
    '''
    Fast impulse response builder. This function takes the image source delays
    and amplitude and fills the impulse response. It uses a linear interpolation
    of the sinc function for speed.

    Parameters
    ----------
    rir: ndarray (double)
        The array to receive the impulse response. It should be filled with
        zero and of the correct size
    time: ndarray (double)
        The array of delays for the image sources
    alpha: ndarray (double)
        The array of attenuations for the image sources
    visibility: ndarray (int)
        Contains 1 if the image source is visible, 0 if not
    fs: int
        The sampling frequency
    fdl: int
        The length of the fractional delay filter (should be odd)
    lut_gran: int
        The number of point per unit in the sinc interpolation table
    '''

    fdl2 = (fdl - 1) // 2
    n_times = time.shape[0]

    assert time.shape[0] == visibility.shape[0]
    assert time.shape[0] == alpha.shape[0]
    assert fdl % 2 == 1

    # check the size of the return array
    max_sample = ceil(fs * np.max(time)) + fdl2
    min_sample = floor(fs * np.min(time)) - fdl2
    assert min_sample >= 0
    assert max_sample < rir.shape[0]

    # create a look-up table of the sinc function and
    # then use linear interpolation
    cdef float delta = 1. / lut_gran
    cdef int lut_size = (fdl + 1) * lut_gran + 1
    n = np.linspace(-fdl2-1, fdl2 + 1, lut_size)

    cdef double [:] sinc_lut = np.sinc(n)
    cdef double [:] hann = np.hanning(fdl)
    cdef int lut_pos, i, f, time_ip
    cdef float x_off, x_off_frac, sample_frac

    for i in range(n_times):
        if visibility[i] == 1:
            # decompose integer and fractional delay
            sample_frac = fs * time[i]
            time_ip = int(floor(sample_frac))
            time_fp = sample_frac - time_ip

            # do the linear interpolation
            x_off_frac = (1. - time_fp) * lut_gran
            lut_gran_off = int(floor(x_off_frac))
            x_off = (x_off_frac - lut_gran_off)
            lut_pos = lut_gran_off
            k = 0
            for f in range(-fdl2, fdl2+1):
                rir[time_ip + f] += alpha[i] * hann[k] * (sinc_lut[lut_pos] 
                        + x_off * (sinc_lut[lut_pos+1] - sinc_lut[lut_pos]))
                lut_pos += lut_gran
                k += 1



from __future__ import division, print_function

from scipy import linalg as la

# import numpy as np

from .doa import *
from .tools_fri_doa_plane import pt_src_recon_multiband, extract_off_diag, cov_mtx_est, polar2cart, make_G, \
    make_GtG_and_inv

class FRIDA(DOA):
    '''
    Implements the FRI-based direction of arrival finding algorithm [FRIDA]_.

    .. note:: Run locate_sources() to apply the CSSM algorithm.

    Parameters
    ----------
    L: ndarray
        Contains the locations of the microphones in the columns
    fs: int or float
        Sampling frequency
    nfft: int
        FFT size
    max_four: int
        Maximum order of the Fourier or spherical harmonics expansion
    c: float, optional
        Speed of sound
    num_src: int, optional
        The number of sources to recover (default 1)
    G_iter: int
        Number of mapping matrix refinement iterations in recovery algorithm (default 1)
    max_ini: int, optional
        Number of random initializations to use in recovery algorithm (default 5)
    n_rot: int, optional
        Number of random rotations to apply before recovery algorithm (default 10)
    noise_level: float, optional
        Noise level in the visibility measurements, if available (default 1e-10)
    stopping: str, optional
        Stopping criteria for the recovery algorithm. Can be max iterations or noise level (default max_iter)
    stft_noise_floor: float
        The noise floor in the STFT measurements, if available (default 0)
    stft_noise_margin: float
        When this, along with stft_noise_floor is set, we only pick frames with at least
        stft_noise_floor * stft_noise_margin power
    signal_type: str
        Which type of measurements to use:

        - 'visibility': Cross correlation measurements
        - 'raw': Microphone signals
    use_lu: bool, optional
        Whether to use LU decomposition for efficiency
    verbose: bool, optional
        Whether to output intermediate result for debugging purposes
    symb: bool, optional
        Whether enforce the symmetry on the reconstructed uniform samples of sinusoids b

    References
    ----------

    .. [FRIDA] H. Pan, R. Scheibler, E. Bezzam, I. Dokmanic, and M. Vetterli, *FRIDA:
        FRI-based DOA estimation for arbitrary array layouts*, Proc. ICASSP,
        pp 3186-3190, 2017
    '''

    def __init__(self, L, fs, nfft, max_four=None, c=343.0, num_src=1,
                 G_iter=None, max_ini=5, n_rot=1, max_iter=50, noise_level=1e-10,
                 low_rank_cleaning=False, stopping='max_iter',
                 stft_noise_floor=0., stft_noise_margin=1.5, signal_type='visibility',
                 use_lu=True, verbose=False, symb=True, use_cache=False, **kwargs):

        DOA.__init__(self, L, fs, nfft, c=c, num_src=num_src, mode='far', **kwargs)

        # intialize some attributes
        self.visi_noisy_all = None
        self.alpha_recon = np.array(num_src, dtype=float)

        # These might be used to select high SNR frames for cross correlation computation
        self.stft_noise_floor = stft_noise_floor
        self.stft_noise_margin = stft_noise_margin

        # Set the number of updates of the mapping matrix
        self.max_four = max_four if max_four is not None else num_src + 1
        self.update_G = True if G_iter is not None and G_iter > 1 else False
        self.G_iter = G_iter if self.update_G else 1
        self.noise_level = noise_level
        self.max_ini = max_ini
        self.max_iter = max_iter
        self.low_rank_cleaning = low_rank_cleaning  # This will impose low rank on corr matrix
        self.stop_cri = stopping
        self.n_rot = n_rot
        self.use_lu = use_lu
        self.verbose = verbose
        self.symb = symb

        self.G = None

        self.mapping_dict = dict(use_cache=use_cache)

        # The type of measurement to use, can be 'visibility' (default) for the covariance
        # matrix, or 'raw' to use microphone signals directly
        self.signal_type = signal_type

    def _process(self, X):
        '''
        Parameters
        ----------
        X: ndarray
            The STFT frames
        '''

        if self.signal_type == 'visibility':
            visi_noisy_all = self._visibilities(X)

            # stack as columns (NOT SUBTRACTING NOISELESS)
            self.visi_noisy_all = np.column_stack(visi_noisy_all)

            signal = self.visi_noisy_all

        elif self.signal_type == 'raw':
            signal = self._raw_average(X)

        else:
            raise ValueError("Signal type can be 'visibility' or 'raw'.")

        # loop over all subbands
        self.num_freq = self.freq_bins.shape[0]

        if self.dim == 2:

            # build the G matrix if necessary
            if self.G is None:
                self.G = make_G(
                    self.L[0, :], self.L[1, :],
                    2 * np.pi * self.freq_hz, self.c,
                    self.max_four,
                    signal_type=self.signal_type
                )
                self.GtG, self.GtG_inv = make_GtG_and_inv(self.G)

            # reconstruct point sources with FRI
            self.azimuth_recon, self.alpha_recon = \
                pt_src_recon_multiband(
                    signal,
                    self.L[0, :], self.L[1, :],
                    2 * np.pi * self.freq_hz, self.c,
                    self.num_src, self.max_four,
                    self.noise_level, self.max_ini,
                    max_iter=self.max_iter,
                    update_G=self.update_G,
                    G_iter=self.G_iter,
                    verbose=False,
                    signal_type=self.signal_type,
                    G_lst=self.G,
                    GtG_lst=self.GtG,
                    GtG_inv_lst=self.GtG_inv
                )

        elif self.dim == 3:

            raise ValueError('3D reconstruction is not yet available with FRIDA.')

    def _raw_average(self, X):
        ''' Correct the time rotation and average the raw microphone signal '''
        phaser = np.exp(-1j * 2 * np.pi * self.freq_hz[:, None] * np.arange(X.shape[2]) * self.nfft / self.fs)

        return np.mean(X[:, self.freq_bins, :] * phaser, axis=2)

    def _visibilities(self, X):

        visi_noisy_all = []
        for band_count in range(self.num_freq):
            # Estimate the covariance matrix and extract off-diagonal entries
            fn = self.freq_bins[band_count]
            energy = np.var(X[:, fn, :], axis=0)
            I = np.where(energy > self.stft_noise_margin * self.stft_noise_floor)
            R = cov_mtx_est(X[:, fn, I[0]])

            # impose low rank constraint
            if self.low_rank_cleaning:
                w, vl = la.eigh(R)
                w = np.abs(w)
                k = self.num_src
                sigma = w.min()  # estimate the noise by minimum statistics
                Rhat = np.dot(vl[:,-k:] * (w[None,-k:] - sigma), np.conj(vl[:,-k:].T))
            else:
                Rhat = R

            visi_noisy = extract_off_diag(Rhat)
            visi_noisy_all.append(visi_noisy)

        return visi_noisy_all

    def _gen_dirty_img(self):
        """
        Compute the dirty image associated with the given measurements. Here the Fourier transform
        that is not measured by the microphone array is taken as zero.
        :return:
        """

        sound_speed = self.c
        num_mic = self.M

        if self.dim == 2:

            x_plt, y_plt = polar2cart(1, self.grid.azimuth)
            img = np.zeros(self.grid.n_points, dtype=complex)

            pos_mic_x = self.L[0, :]
            pos_mic_y = self.L[1, :]
            for i in range(self.num_freq):

                visi = self.visi_noisy_all[:, i]
                omega_band = 2 * np.pi * self.freq_hz[i]

                pos_mic_x_normalised = pos_mic_x / (sound_speed / omega_band)
                pos_mic_y_normalised = pos_mic_y / (sound_speed / omega_band)

                count_visi = 0
                for q in range(num_mic):
                    p_x_outer = pos_mic_x_normalised[q]
                    p_y_outer = pos_mic_y_normalised[q]
                    for qp in range(num_mic):
                        if not q == qp:
                            p_x_qqp = p_x_outer - pos_mic_x_normalised[qp]  # a scalar
                            p_y_qqp = p_y_outer - pos_mic_y_normalised[qp]  # a scalar
                            # <= the negative sign converts DOA to propagation vector
                            img += visi[count_visi] * \
                                   np.exp(-1j * (p_x_qqp * x_plt + p_y_qqp * y_plt))
                            count_visi += 1

            return img / (num_mic * (num_mic - 1))

        elif self.dim == 3:

            raise ValueError('3D reconstruction is not yet available with FRIDA.')

# -------------MISC--------------#

# def polar2cart(rho, phi):
#     """
#     convert from polar to cartesian coordinates
#     :param rho: radius
#     :param phi: azimuth
#     :return:
#     """
#     x = rho * np.cos(phi)
#     y = rho * np.sin(phi)
#     return x, y



"""
Single Channel Noise Reduction
==============================

Collection of single channel noise reduction (SCNR) algorithms for speech:

- :doc:`Spectral Subtraction <pyroomacoustics.denoise.spectral_subtraction>` [1]_
- :doc:`Subspace Approach <pyroomacoustics.denoise.subspace>` [2]_
- :doc:`Iterative Wiener Filtering <pyroomacoustics.denoise.iterative_wiener>` [3]_

At `this repository <https://github.com/santi-pdp/segan>`_, a deep learning approach in Python can be found.

References
----------

.. [1] M. Berouti, R. Schwartz, and J. Makhoul, *Enhancement of speech corrupted by acoustic noise,*
    ICASSP '79. IEEE International Conference on Acoustics, Speech, and Signal Processing, 1979, pp. 208-211.

.. [2] Y. Ephraim and H. L. Van Trees, *A signal subspace approach for speech enhancement,*
    IEEE Transactions on Speech and Audio Processing, vol. 3, no. 4, pp. 251-266, Jul 1995.

.. [3] J. Lim and A. Oppenheim, *All-Pole Modeling of Degraded Speech,*
    IEEE Transactions on Acoustics, Speech, and Signal Processing 26.3 (1978): 197-210.

"""

from .spectral_subtraction import *
from .subspace import *
from .iterative_wiener import *



# Utility for downloading and uncompressing dataset
# Copyright (C) 2019 Robin Scheibler, MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.

import os
import tarfile
import bz2

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen


def download_uncompress(url, path='.', compression=None, context=None):
    '''
    This functions download and uncompress on the fly a file
    of type tar, tar.gz, tar.bz2.

    Parameters
    ----------
    url: str
        The URL of the file
    path: str, optional
        The path where to uncompress the file
    compression: str, optional
        The compression type (one of 'bz2', 'gz', 'tar'), infered from url
        if not provided
    context: SSL certification, optional
        Default is to use none.
    '''

    # infer compression from url
    if compression is None:
        compression = os.path.splitext(url)[1][1:]

    # check compression format and set mode
    if compression in ['gz', 'bz2']:
        mode = 'r|' + compression
    elif compression == 'tar':
        mode = 'r:'
    else:
        raise ValueError('The file must be of type tar/gz/bz2.')

    # download and untar/uncompress at the same time
    if context is not None:
        stream = urlopen(url, context=context)
    else:
        stream = urlopen(url)
    tf = tarfile.open(fileobj=stream, mode=mode)
    tf.extractall(path)



from __future__ import division, print_function
from unittest import TestCase
import numpy as np
import os
from pyroomacoustics import create_noisy_signal, rms, normalize
from scipy.io import wavfile

tol = 1e-5

signal_fp = os.path.join(os.path.dirname(__file__), '..', '..', 'examples',
                         'input_samples', 'cmu_arctic_us_aew_a0001.wav')

def white_noise(snr):
    np.random.seed(0)
    noisy_signal, signal, noise, fs = create_noisy_signal(signal_fp, snr=snr)
    _snr = 20 * np.log10(rms(signal) / rms(noise))
    err = abs(snr-_snr)
    return err, noisy_signal, fs


class TestCreateNoisySignal(TestCase):

    def test_snr_error(self):
        snr_test = 0
        res = white_noise(snr_test)
        self.assertTrue(res[0] < tol)


if __name__ == "__main__":
    snr = 0
    err, noisy_signal, fs = white_noise(snr)
    print("SNR error: {}".format(err))

    # write to WAV, need to cast to `np.float32` for float values
    # see table here: https://docs.scipy.org/doc/scipy/reference/generated/scipy.io.wavfile.write.html
    wavfile.write("noisy_signal_{}dB.wav".format(snr), fs,
                  noisy_signal.astype(np.float32))



# coding=utf-8
# Phase reconstruction with the Griffin-Lim algorithm
#
# Copyright (C) 2019 Robin Scheibler, MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# You should have received a copy of the MIT License along with this program. If
# not, see <https://opensource.org/licenses/MIT>.
"""
Implementation of the classic phase reconstruction from Griffin and Lim [1]_.
The input to the algorithm is the magnitude from STFT measurements.

The algorithm works by starting by assigning a (possibly random) initial phase to the
measurements, and then iteratively

1. Reconstruct the time-domain signal
2. Re-apply STFT
3. Enforce the known magnitude of the measurements

The implementation supports different types of initialization via the keyword argument ``ini``.

1. If omitted, the initial phase is uniformly zero
2. If ``ini="random"``, a random phase is used
3. If ``ini=A`` for a ``numpy.ndarray`` of the same shape as the input magnitude, ``A / numpy.abs(A)`` is used for initialization

Example
-------

.. code-block:: python

    import numpy as np
    from scipy.io import wavfile
    import pyroomacoustics as pra

    # We open a speech sample
    filename = "examples/input_samples/cmu_arctic_us_axb_a0004.wav"
    fs, audio = wavfile.read(filename)

    # These are the parameters of the STFT
    fft_size = 512
    hop = fft_size // 4
    win_a = np.hamming(fft_size)
    win_s = pra.transform.compute_synthesis_window(win_a, hop)
    n_iter = 200

    engine = pra.transform.STFT(
        fft_size, hop=hop, analysis_window=win_a, synthesis_window=win_s
    )
    X = engine.analysis(audio)
    X_mag = np.abs(X)
    X_mag_norm = np.linalg.norm(X_mag) ** 2

    # monitor convergence
    errors = []

    # the callback to track the spectral distance convergence
    def cb(epoch, Y, y):
        # we measure convergence via spectral distance
        Y_2 = engine.analysis(y)
        sd = np.linalg.norm(X_mag - np.abs(Y_2)) ** 2 / X_mag_norm
        # save in the list every 10 iterations
        if epoch % 10 == 0:
            errors.append(sd)

    pra.phase.griffin_lim(X_mag, hop, win_a, n_iter=n_iter, callback=cb)

    plt.semilogy(np.arange(len(errors)) * 10, errors)
    plt.show()


References
----------

.. [1] D. Griffin and J. Lim, “Signal estimation from modified short-time Fourier
    transform,” IEEE Transactions on Acoustics, Speech, and Signal Processing, vol.
    32, no. 2, pp. 236–243, 1984.
"""
# This is needed to check for string types
# in a way compatible between python 2 and 3
try:
    basestring
except NameError:
    basestring = str

import numpy as np
from ..transform import STFT, compute_synthesis_window


def griffin_lim(
    X,
    hop,
    analysis_window,
    fft_size=None,
    stft_kwargs={},
    n_iter=100,
    ini=None,
    callback=None,
):
    """
    Implementation of the Griffin-Lim phase reconstruction algorithm from STFT magnitude measurements.

    Parameters
    ----------
    X: array_like, shape (n_frames, n_freq)
        The STFT magnitude measurements
    hop: int
        The frame shift of the STFT
    analysis_window: array_like, shape (fft_size,)
        The window used for the STFT analysis
    fft_size: int, optional
        The FFT size for the STFT, if omitted it is computed from the dimension of ``X``
    stft_kwargs: dict, optional
        Dictionary of extra parameters for the STFT
    n_iter: int, optional
        The number of iteration
    ini: str or array_like, np.complex, shape (n_frames, n_freq), optional
        The initial value of the phase estimate. If "random", uses a random guess. If ``None``, uses ``0`` phase.
    callback: func, optional
        A callable taking as argument an int and the reconstructed STFT and time-domain signals
    """

    if isinstance(ini, basestring) and ini == "random":
        ini = np.exp(1j * 2 * np.pi * np.random.rand(*X.shape))
    elif ini is None:
        ini = np.ones(X.shape, dtype=np.complex128)
    else:
        # make sure the modulus is one
        ini /= np.abs(ini)

    # take care of the STFT parameters
    if fft_size is None:
        fft_size = 2 * (X.shape[1] - 1)

    # the optimal GL window
    synthesis_window = compute_synthesis_window(analysis_window, hop)

    # create the STFT object
    engine = STFT(
        fft_size,
        hop=hop,
        analysis_window=analysis_window,
        synthesis_window=synthesis_window,
        **stft_kwargs
    )

    # Initialize the signal
    Y = X * ini
    y = engine.synthesis(Y)

    # the successive application of analysis/synthesis introduces
    # a shift of ``fft_size - hop`` that we must correct
    the_shift = fft_size - hop
    y[:-the_shift,] = y[the_shift:,]

    for epoch in range(n_iter):

        # possibly monitor the reconstruction
        if callback is not None:
            callback(epoch, Y, y)

        # back to STFT domain
        Y[:, :] = engine.analysis(y)

        # enforce magnitudes
        Y *= X / np.abs(Y)

        # back to time domain
        y[:-the_shift,] = engine.synthesis(Y)[the_shift:,]

    # last callback
    if callback is not None:
        callback(epoch, Y, y)

    return y




from __future__ import division, print_function
import numpy as np
from scipy import fftpack

import scipy.linalg as la
try:
    import mkl_fft as fft
    has_mklfft = True
except ImportError:
    import numpy.fft as fft
    has_mklfft = False

def autocorr(x):
    """ Fast autocorrelation computation using the FFT """
    
    X = fft.rfft(x, n=2*x.shape[0])
    r = fft.irfft(np.abs(X)**2)
    
    return r[:x.shape[0]] / (x.shape[0] - 1)

def toeplitz_multiplication(c, r, A, **kwargs):
    """ 
    Compute numpy.dot(scipy.linalg.toeplitz(c,r), A) using the FFT.

    Parameters
    ----------
    c: ndarray
        the first column of the Toeplitz matrix
    r: ndarray
        the first row of the Toeplitz matrix
    A: ndarray
        the matrix to multiply on the right
    """
    
    m = c.shape[0]
    n = r.shape[0]

    fft_len = int(2**np.ceil(np.log2(m+n-1)))
    zp = fft_len - m - n + 1
    
    if A.shape[0] != n:
        raise ValueError('A dimensions not compatible with toeplitz(c,r)')
    
    x = np.concatenate((c, np.zeros(zp, dtype=c.dtype), r[-1:0:-1]))
    xf = np.fft.rfft(x, n=fft_len)
    
    Af = np.fft.rfft(A, n=fft_len, axis=0)
    
    return np.fft.irfft((Af.T*xf).T, n=fft_len, axis=0)[:m,]

def hankel_multiplication(c, r, A, mkl=True, **kwargs):
    '''
    Compute numpy.dot(scipy.linalg.hankel(c,r=r), A) using the FFT.

    Parameters
    ----------
    c: ndarray
        the first column of the Hankel matrix
    r: ndarray
        the last row of the Hankel matrix
    A: ndarray
        the matrix to multiply on the right
    mkl: bool, optional
        if True, use the mkl_fft package if available
    '''
    
    if mkl and has_mklfft:
        fmul = mkl_toeplitz_multiplication
    else:
        fmul = toeplitz_multiplication
        A = A[:r.shape[0],:]

    return fmul(c[::-1], r, A, **kwargs)[::-1,]


def mkl_toeplitz_multiplication(c, r, A, A_padded=False, out=None, fft_len=None):
    """ 
    Compute numpy.dot(scipy.linalg.toeplitz(c,r), A) using the FFT from the mkl_fft package.

    Parameters
    ----------
    c: ndarray
        the first column of the Toeplitz matrix
    r: ndarray
        the first row of the Toeplitz matrix
    A: ndarray
        the matrix to multiply on the right
    A_padded: bool, optional
        the A matrix can be pre-padded with zeros by the user, if this is the case
        set to True
    out: ndarray, optional
        an ndarray to store the output of the multiplication
    fft_len: int, optional
        specify the length of the FFT to use
    """

    if not has_mklfft:
        raise ValueError('Import mkl_fft package unavailable. Install from https://github.com/LCAV/mkl_fft')

    m = c.shape[0]
    n = r.shape[0]

    if fft_len is None:
        fft_len = int(2**np.ceil(np.log2(m+n-1)))
    zp = fft_len - m - n + 1
    
    if (not A_padded and A.shape[0] != n) or (A_padded and A.shape[0] != fft_len):
        raise ValueError('A dimensions not compatible with toeplitz(c,r)')
    
    x = np.concatenate((c, np.zeros(zp, dtype=c.dtype), r[-1:0:-1]))
    xf = fft.rfft(x, n=fft_len)
    
    if out is not None:
        fft.rfft(A, n=fft_len, axis=0, out=out)
    else:
        out = fft.rfft(A, n=fft_len, axis=0)

    out *= xf[:,None]

    if A_padded:
        fft.irfft(out, n=fft_len, axis=0, out=A)
    else:
        A = fft.irfft(out, n=fft_len, axis=0)
    
    return A[:m,]


def naive_toeplitz_multiplication(c, r, A):
    """ 
    Compute numpy.dot(scipy.linalg.toeplitz(c,r), A)

    Parameters
    ----------
    c: ndarray
        the first column of the Toeplitz matrix
    r: ndarray
        the first row of the Toeplitz matrix
    A: ndarray
        the matrix to multiply on the right
    """

    return np.dot(la.toeplitz(c,r),A)

def hankel_stride_trick(x, shape):
    ''' 
    Make a Hankel matrix from a vector using stride tricks 
    
    Parameters
    ----------
    x: ndarray
        a vector that contains the concatenation of the first column
        and first row of the Hankel matrix to build *without* repetition
        of the lower left corner value of the matrix
    shape: tuple
        the shape of the Hankel matrix to build, it must satisfy ``x.shape[0] == shape[0] + shape[1] - 1``
    '''

    if x.shape[0] != shape[0] + shape[1] - 1:
        raise ValueError('Inconsistent dimensions')

    strides = (x.itemsize, x.itemsize)
    return np.lib.stride_tricks.as_strided(x, shape=shape, strides=strides)

def toeplitz_strang_circ_approx(r, matrix=False):
    '''
    Circulant approximation to a symetric Toeplitz matrix
    by Gil Strang

    Parameters
    ----------
    r: ndarray
        the first row of the symmetric Toeplitz matrix
    matrix: bool, optional
        if True, the full symetric Toeplitz matrix is returned,
        otherwise, only the first column
    '''

    n = r.shape[0]
    c = r.copy()
    m = n // 2 if n % 2 == 0 else (n - 1) // 2
    c[-m:] = r[m:0:-1]

    if matrix:
        return la.circulant(c)
    else:
        return c
    

def toeplitz_opt_circ_approx(r, matrix=False):
    ''' 
    Optimal circulant approximation of a symmetric Toeplitz matrix
    by Tony F. Chan

    Parameters
    ----------
    r: ndarray
        the first row of the symmetric Toeplitz matrix
    matrix: bool, optional
        if True, the full symetric Toeplitz matrix is returned,
        otherwise, only the first column
    '''

    n = r.shape[0]

    r_rev = np.zeros(r.shape)
    r_rev[1:] = r[:0:-1]

    i = np.arange(n)
    c = (i * r_rev + (n - i) * r) / n
    c[1:] = c[:0:-1]

    if matrix:
        return la.circulant(c)
    else:
        return c


if __name__ == "__main__":

    import time 

    try:
        import mkl as mkl_service
        mkl_service.set_num_threads(7)
    except ImportError:
        pass

    n_iter = 1

    m = 10
    n = 100
    o = 10

    dtype_in = np.float32
    dtype_out = np.complex128 if dtype_in == np.float64 else np.complex64
    dorder = 'C'
    fft_len = int(2 ** np.ceil( np.log2(m + n - 1) ))

    c = dtype_in(np.random.randn(m))
    r = dtype_in(np.random.randn(n))

    A = np.asfortranarray(dtype_in(np.random.randn(n, o)))

    Apad = np.zeros((fft_len, o), order=dorder, dtype=dtype_in)
    Apad[:n,:] = A
    out = np.zeros((fft_len//2+1, o), dtype=dtype_out, order=dorder)

    print('Start...')
    start_time = time.time()
    for i in range(n_iter):
        C1 = naive_toeplitz_multiplication(c, r, A)
    nai_time = time.time() - start_time
    print("naive --- %s seconds ---" % (nai_time))

    '''
    start_time = time.time()
    for i in range(n_iter):
        C = toeplitz_multiplication(c, r, A)
    npy_time = time.time() - start_time
    print("numpy   --- %s seconds ---" % (npy_time))
    '''

    start_time = time.time()
    for i in range(n_iter):
        #C2 = mkl_toeplitz_multiplication(c, r, Apad, A_padded=True, out=out, fft_len=fft_len)
        C2 = mkl_toeplitz_multiplication(c, r, A, A_padded=False, out=None, fft_len=None)
    mkl_time = time.time() - start_time

    print("mkl     --- %s seconds ---" % (mkl_time))
    print("Result matching:", np.allclose(C1, C2))

    print('Stop.')



# encoding: utf-8
'''
Phase Processing
================

This sub-package contains algorithms related to phase specific process, such as phase reconstruction.

Griffin-Lim
    | Phase reconstruction by fixed-point iterations [1]_
    | :py:mod:`pyroomacoustics.phase.gl`

References
----------

.. [1] D. Griffin and J. Lim, “Signal estimation from modified short-time Fourier
    transform,” IEEE Transactions on Acoustics, Speech, and Signal Processing, vol.
    32, no. 2, pp. 236–243, 1984.

'''

from .gl import griffin_lim



# -*- coding: utf-8 -*-
'''
Blind Source Separation
=======================

Implementations of a few blind source separation (BSS) algorithms.

AuxIVA
    | Independent Vector Analysis [1]_
    | :py:mod:`pyroomacoustics.bss.auxiva`
Trinicon
    | Time-domain BSS [2]_
    | :py:mod:`pyroomacoustics.bss.trinicon`
ILRMA
    | Independent Low-Rank Matrix Analysis [3]_
    | :py:mod:`pyroomacoustics.bss.ilrma`
SparseAuxIVA
    | Sparse Independent Vector Analysis [4]_
    | :py:mod:`pyroomacoustics.bss.sparseauxiva`
FastMNMF
    | Fast Multichannel Nonnegative Matrix Factorization [5]_
    | :py:mod:`pyroomacoustics.bss.fastmnmf`


A few commonly used functions, such as projection back, can be found in
:py:mod:`pyroomacoustics.bss.common`.

References
----------

.. [1] N. Ono, *Stable and fast update rules for independent vector analysis based
    on auxiliary function technique,* Proc. IEEE, WASPAA, pp. 189-192, Oct. 2011.


.. [2] R. Aichner, H. Buchner, F. Yan, and W. Kellermann  *A real-time
    blind source separation scheme and its application to reverberant and noisy
    acoustic environments*,  Signal Processing, 86(6), 1260-1277.
    doi:10.1016/j.sigpro.2005.06.022, 2006.

.. [3] D. Kitamura, N. Ono, H. Sawada, H. Kameoka, H. Saruwatari, *Determined blind
    source separation unifying independent vector analysis and nonnegative matrix
    factorization,* IEEE/ACM Trans. ASLP, vol. 24, no. 9, pp. 1626-1641, Sept. 2016

.. [4] J. Janský, Z. Koldovský, and N. Ono, *A computationally cheaper method
    for blind speech separation based on AuxIVA and incomplete demixing transform,*
    Proc. IEEE, IWAENC, pp. 1-5, Sept. 2016.

.. [5] K. Sekiguchi, A. A. Nugraha, Y. Bando, K. Yoshii, *Fast Multichannel Source 
    Separation Based on Jointly Diagonalizable Spatial Covariance Matrices*, EUSIPCO, 2019.

'''

from .trinicon import trinicon
from .auxiva import auxiva
from .ilrma import ilrma
from .sparseauxiva import sparseauxiva
from .fastmnmf import fastmnmf
from .common import projection_back, sparir



'''
The Datasets Sub-package is responsible to deliver
wrappers around a few popular audio datasets to make
them easier to use.

Two base class :py:obj:`pyroomacoustics.datasets.base.Dataset` and
:py:obj:`pyroomacoustics.datasets.base.Sample` wrap
together the audio samples and their meta data.
The general idea is to create a sample object with an attribute
containing all metadata. Dataset objects that have a collection
of samples can then be created and can be filtered according
to the values in the metadata.

Many of the functions with ``match`` or ``filter`` will take an
arbitrary number of keyword arguments. The keys should match some
metadata in the samples. Then there are three ways that match occurs
between a ``key/value`` pair and an ``attribute`` sharing the same key.

1. ``value == attribute``
2. ``value`` is a list and ``attribute in value == True``
3. ``value`` is a callable (a function) and ``value(attribute) == True``

Example 1
---------

::

    # Prepare a few artificial samples
    samples = [
        {
            'data' : 0.99,
            'metadata' : { 'speaker' : 'alice', 'sex' : 'female', 'age' : 37, 'number' : 'one' },
        },
        {
            'data' : 2.1,
            'metadata' : { 'speaker' : 'alice', 'sex' : 'female', 'age' : 37, 'number' : 'two' },
        },
        {
            'data' : 1.02,
            'metadata' : { 'speaker' : 'bob', 'sex' : 'male', 'age' : 48, 'number' : 'one' },
        },
        {
            'data' : 2.07,
            'metadata' : { 'speaker' : 'bob', 'sex' : 'male', 'age' : 48, 'number' : 'two' },
        },
        ]

    corpus = Dataset()
    for s in samples:
        new_sample = Sample(s['data'], **s['metadata'])
        corpus.add_sample(new_sample)

    # Then, it possible to display summary info about the corpus
    print(corpus)

    # The number of samples in the corpus is given by ``len``
    print('Number of samples:', len(corpus))

    # And we can access samples with the slice operator
    print('Sample #2:')
    print(corpus[2])    # (shortcut for `corpus.samples[2]`)

    # We can obtain a new corpus with only male subject
    corpus_male_only = corpus.filter(sex='male')
    print(corpus_male_only)

    # Only retain speakers above 40 years old
    corpus_older = corpus.filter(age=lambda a : a > 40)
    print(corpus_older)

Example 2 (CMU ARCTIC)
----------------------

::

    # This example involves the CMU ARCTIC corpus available at
    # http://www.festvox.org/cmu_arctic/

    import matplotlib.pyplot as plt
    import pyroomacoustics as pra

    # Here, the corpus for speaker bdl is automatically downloaded
    # if it is not available already
    corpus = pra.datasets.CMUArcticCorpus(download=True, speaker=['bdl'])

    # print dataset info and 10 sentences
    print(corpus)
    corpus.head(n=10)

    # let's extract all samples containing the word 'what'
    keyword = 'what'
    matches = corpus.filter(text=lambda t : keyword in t)
    print('The number of sentences containing "{}": {}'.format(keyword, len(matches)))
    for s in matches.sentences:
        print('  *', s)

    # if the sounddevice package is available, we can play the sample
    matches[0].play()

    # show the spectrogram
    plt.figure()
    matches[0].plot()
    plt.show()

Example 3 (Google's Speech Commands Dataset)
--------------------------------------------

::

    # This example involves Google's Speech Commands Dataset available at
    # https://research.googleblog.com/2017/08/launching-speech-commands-dataset.html

    import matplotlib.pyplot as plt
    import pyroomacoustics as pra

    # The dataset is automatically downloaded if not available and 10 of each word is selected
    dataset = pra.datasets.GoogleSpeechCommands(download=True, subset=10, seed=0)

    # print dataset info, first 10 entries, and all sounds
    print(dataset)
    dataset.head(n=10)
    print("All sounds in the dataset:")
    print(dataset.classes)

    # filter by specific word
    selected_word = 'yes'
    matches = dataset.filter(word=selected_word)
    print("Number of '%s' samples : %d" % (selected_word, len(matches)))

    # if the sounddevice package is available, we can play the sample
    matches[0].play()

    # show the spectrogram
    plt.figure()
    matches[0].plot()
    plt.show()
  
'''

from .base import Meta, Sample, AudioSample, Dataset
from .timit import Word, Sentence, TimitCorpus
from .cmu_arctic import CMUArcticCorpus, CMUArcticSentence, cmu_arctic_speakers
from .google_speech_commands import GoogleSpeechCommands, GoogleSample



'''
This tests the construction of a bank of octave filters
'''
import pyroomacoustics as pra

import numpy as np
from scipy.signal import sosfreqz

tol = 1.  # decibel

def test_bandpass_filterbank():
    f0 = 125. / 2.
    fs = 16000

    bands, fc = pra.octave_bands(f0)
    filters = pra.bandpass_filterbank(bands, fs=fs, order=16)

    res = None
    for sos, f in zip(filters, fc):
        w, h = sosfreqz(sos, worN=4096, whole=False)

        # inspect the cumulative response
        if res is None:
            res = np.abs(h) ** 2
        else:
            res += np.abs(h) ** 2

    # check automatically that the cumulative response is (nearly) flat
    freq = w / np.pi * fs / 2
    I = np.where(np.logical_and(freq > fc[0], freq < np.minimum(fc[-1], 0.95 * fs / 2)))[0]
    err = np.max(np.abs(10 * np.log10(res[I])))
    
    assert err <= tol, 'Bandpass filterbank test fails (err {} > tol {})'.format(err, tol)


if __name__ == '__main__':

    import matplotlib.pyplot as plt

    f0 = 125. / 2.
    fs = 16000

    bands, fc = pra.octave_bands(f0)
    filters = pra.bandpass_filterbank(bands, fs=fs, order=16)

    plt.figure()
    res = None
    for sos, f in zip(filters, fc):
        w, h = sosfreqz(sos, worN=4096, whole=False)
        plt.semilogx(w / np.pi * fs / 2., 20 * np.log10(np.abs(h)), label=str(f) + ' Hz')

        # inspect the cumulative response
        if res is None:
            res = np.abs(h) ** 2
        else:
            res += np.abs(h) ** 2

    # check automatically that the cumulative response is (nearly) flat
    freq = w / np.pi * fs / 2
    I = np.where(np.logical_and(freq > fc[0], freq < np.minimum(fc[-1], 0.95 * fs / 2)))[0]
    err = np.max(np.abs(10 * np.log10(res[I])))
    print('The error is', err, '(tol is', tol, 'dB)')

    plt.semilogx(w / np.pi * fs / 2., 10 * np.log10(np.abs(res)), label='Sum')

    plt.title('Octave Filter Bank')
    plt.xlabel('Frequency [Hz]')
    plt.ylabel('Amplitude [dB]')
    plt.margins(0, 0.1)
    plt.grid(which='both', axis='both')
    plt.axvline(100, color='green') # cutoff frequency
    plt.ylim(-60,3)
    plt.xlim(10., fs / 2.)
    plt.legend()
    plt.show()

    # run the automatic tests
    test_bandpass_filterbank()




from __future__ import division, print_function
import numpy as np
import pyroomacoustics as pra
import time

try:
    import pyfftw
    pyfftw_available = True
except ImportError:
    pyfftw_available = False

try:
    import mkl_fft
    mkl_available = True
except ImportError:
    mkl_available = False


n_trials = 1000
nfft = 128
D = 7
x = np.random.randn(nfft, D).astype('float32')

def timing(transform, n_trials):

    dft = pra.transform.DFT(nfft, D, transform=transform)
    start_time = time.time()
    for k in range(n_trials):
        dft.analysis(x)
    analysis_time = (time.time()-start_time)/n_trials * 1e6
    start_time = time.time()
    for k in range(n_trials):
        dft.synthesis()
    synthesis_time = (time.time()-start_time)/n_trials * 1e6

    print("avg %s : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]" % 
        (transform, analysis_time+synthesis_time, analysis_time, synthesis_time))


res = timing('numpy', n_trials)
if pyfftw_available:
    res = timing('fftw', n_trials)
if mkl_available:
    res = timing('mkl', n_trials)

"""
test against without using class
"""
print()
start_time = time.time()
for k in range(n_trials):
    X = np.fft.rfft(x)
analysis_time = (time.time()-start_time)/n_trials * 1e6
start_time = time.time()
for k in range(n_trials):
    x_r = np.fft.irfft(X)
synthesis_time = (time.time()-start_time)/n_trials * 1e6
print("avg numpy w/o class : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]" % 
        (analysis_time+synthesis_time, analysis_time, synthesis_time))

if pyfftw_available:

    # prepare
    a = pyfftw.empty_aligned([nfft, D], dtype='float32')
    b = pyfftw.empty_aligned([nfft//2+1, D], dtype='complex64')
    c = pyfftw.empty_aligned([nfft, D], dtype='float32')
    forward = pyfftw.FFTW(a, b, axes=(0, ))
    backward = pyfftw.FFTW(b, c, axes=(0, ), direction='FFTW_BACKWARD')

    start_time = time.time()
    for k in range(n_trials):
        forward()
    analysis_time = (time.time()-start_time)/n_trials * 1e6
    start_time = time.time()
    for k in range(n_trials):
        backward()
    synthesis_time = (time.time()-start_time)/n_trials * 1e6
    print("avg fftw w/o class : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]" % 
            (analysis_time+synthesis_time, analysis_time, synthesis_time))

if mkl_available:
    start_time = time.time()
    for k in range(n_trials):
        X = mkl_fft.rfft_numpy(x)
    analysis_time = (time.time()-start_time)/n_trials * 1e6
    start_time = time.time()
    for k in range(n_trials):
        x_r = mkl_fft.irfft_numpy(X)
    synthesis_time = (time.time()-start_time)/n_trials * 1e6
    print("avg mkl w/o class : %f [1e-6 sec], (analysis, synthesis)=(%f, %f) [1e-6 sec]" % 
            (analysis_time+synthesis_time, analysis_time, synthesis_time))
    



from __future__ import division, print_function

from unittest import TestCase
import numpy as np
from scipy.signal import fftconvolve
import pyroomacoustics as pra

# fix RNG for a deterministic result
np.random.seed(0)

# parameters
num_taps = 6        # the unknown filter length
n_samples = 20000   # the number of samples to run
tol = 1e-5         # tolerance reconstructed filter
fft_length = 128   # block size

# the unknown filters in the frequency domain
num_bands = fft_length//2+1
W = np.random.randn(num_taps,num_bands) + \
    1j*np.random.randn(num_taps,num_bands)
W /= np.linalg.norm(W, axis=0)

# create a known driving signal
x = np.random.randn(n_samples)

# take to STFT domain
window = pra.hann(fft_length)  # the analysis window
hop = fft_length//2
stft_in = pra.transform.STFT(fft_length, hop=hop,
                             analysis_window=window)

n = 0
num_blocks = 0
X_concat = np.zeros((num_bands,n_samples//hop),dtype=np.complex64)
while  n_samples - n > hop:

    stft_in.analysis(x[n:n+hop,])
    X_concat[:,num_blocks] = stft_in.X

    n += hop
    num_blocks += 1

# convolve in frequency domain with unknown filter
Y_concat = np.zeros((num_bands,num_blocks), dtype=np.complex64)
for k in range(num_bands):
    Y_concat[k,:] = fftconvolve(X_concat[k,:], W[:,k])[:num_blocks]

# run filters on each block
def run_filters(algorithm, X_concat, Y_concat):
    num_blocks = X_concat.shape[1]
    for n in range(num_blocks):
        algorithm.update(X_concat[:,n], Y_concat[:,n])


class TestAdaptiveFilterFrequencyDomain(TestCase):

    def test_subband_nlms(self):

        subband_nlms = pra.adaptive.SubbandLMS(num_taps=num_taps, 
            num_bands=num_bands, mu=0.5, nlms=True)
        run_filters(subband_nlms, X_concat, Y_concat)
        error_per_band = np.linalg.norm(subband_nlms.W.conj() - W, axis=0)
        error = np.max(abs(error_per_band))
        print('Subband NLMS Reconstruction Error', error)
        self.assertTrue(error < tol)







"""
Just run the Material command with a bunch of inputs to make sure
it works as expected
"""
import pyroomacoustics as pra


scat_test = {
    "coeffs": [0.1, 0.1, 0.1, 0.2, 0.2, 0.2, 0.3],
    "center_freqs": [125, 250, 500, 1000, 2000, 4000, 8000],
}
abs_test = {
    "coeffs": [0.3, 0.4, 0.25, 0.11, 0.05, 0.03, 0.3],
    "center_freqs": [125, 250, 500, 1000, 2000, 4000, 8000],
}


def test_material_e_float():
    mat = pra.Material(0.3)


def test_material_es_float():
    mat = pra.Material(0.3, 0.1)


def test_material_e_db():
    mat = pra.Material("hard_surface")


def test_material_es_db():
    mat = pra.Material("hard_surface", "rpg_qrd")


def test_material_e_dict():
    mat = pra.Material(abs_test)


def test_material_es_dict():
    mat = pra.Material(abs_test, scat_test)


def test_material_e_float_s_db():
    mat = pra.Material(0.3, "rpg_skyline")


def test_material_e_db_s_float():
    mat = pra.Material("brickwork", 0.1)


def test_material_e_float_s_dict():
    mat = pra.Material(0.1, scat_test)


def test_material_e_dict_s_float():
    mat = pra.Material(abs_test, 0.1)


def test_material_e_db_s_dict():
    mat = pra.Material("brickwork", scat_test)


def test_material_e_dict_s_db():
    mat = pra.Material(abs_test, "classroom_tables")


def test_dict_pairs():
    materials = pra.make_materials(
        ceiling=(0.25, 0.01),
        floor=(0.5, 0.1),
        east=(0.15, 0.15),
        west=(0.07, 0.15),
        north=(0.15, 0.15),
        south=(0.10, 0.15),
    )

    assert isinstance(materials, dict)


def test_list_pairs():
    materials = pra.make_materials(
        (0.25, 0.01),
        (0.5, 0.1),
        (0.15, 0.15),
        (0.07, 0.15),
        (0.15, 0.15),
        (0.10, 0.15),
    )

    assert isinstance(materials, list)


def test_dict_list_mix():
    mat_list, mat_dict = pra.make_materials(
        (0.25, 0.01),
        abs_test,
        "brickwork",
        ("brickwork", scat_test),
        1.,
        ceilling=(abs_test, scat_test),
        floor=(abs_test, "rpg_skyline"),
        one=(0.10, 0.15),
    )

    assert isinstance(mat_list, list)
    assert isinstance(mat_dict, dict)


def test_empty():
    mat_list = pra.make_materials()

    assert mat_list == []


if __name__ == '__main__':

    test_material_e_float()
    test_material_es_float()
    test_material_es_dict
    test_material_e_db()
    test_material_es_db()
    test_material_e_dict()
    test_material_es_dict()
    test_material_e_float_s_db()
    test_material_e_db_s_float()
    test_material_e_float_s_dict()
    test_material_e_dict_s_float()
    test_material_e_db_s_dict()
    test_material_e_dict_s_db()
    test_dict_pairs()
    test_list_pairs()
    test_dict_list_mix()
    test_empty()



# @version: 1.0  date: 05/06/2015 by Sidney Barthe
# @author: robin.scheibler@epfl.ch, ivan.dokmanic@epfl.ch, sidney.barthe@epfl.ch
# @copyright: EPFL-IC-LCAV 2015

import unittest
import numpy as np

import pyroomacoustics as pra

fs = 8000
t0 = 0.
absorption = 0.90
max_order_sim = 2
sigma2_n = 5e-7

corners = np.array([[0,0,6,6],[0,6,6,0]])
room = pra.Room.from_corners(
    corners,
    absorption=absorption,
    fs=fs,
    t0=t0,
    max_order=max_order_sim,
    sigma2_awgn=sigma2_n)

room.extrude(6)

room.add_source([3, 3, 3], None, 0)

# place 3 microphones in the room
mics = pra.MicrophoneArray(np.array([[3, 5, 0, 7],
                                     [3, 3, 0, 7],
                                     [3, 3, 0, 7]]), fs)
room.add_microphone_array(mics)

# run the image source model
room.image_source_model()

# we order the sources lexicographically
ordering = np.lexsort(room.sources[0].images)
images_found = room.sources[0].images[:,ordering]
visibilities = np.array(room.visibility[0][:,ordering], dtype=bool)

sources = np.array([
    [  3.,   3.,   3.,  -3.,  -3.,   3.,   9.,   9.,   3.,   3.,   3.,
        -3.,  -3.,   3.,   9.,   9.,  -9.,  -3.,   3.,   9.,  15.,  -3.,
        -3.,   3.,   9.,   9.,   3.,   3.,   3.,  -3.,  -3.,   3.,   9.,
        9.,   3.,   3.,   3.],
    [  3.,  -3.,  -3.,   3.,   3.,   3.,   3.,   3.,   9.,   9.,  -9.,
        -3.,  -3.,  -3.,  -3.,  -3.,   3.,   3.,   3.,   3.,   3.,   9.,
        9.,   9.,   9.,   9.,  15.,  -3.,  -3.,   3.,   3.,   3.,   3.,
        3.,   9.,   9.,   3.],
    [ -9.,  -3.,  -3.,  -3.,  -3.,  -3.,  -3.,  -3.,  -3.,  -3.,   3.,
        3.,   3.,   3.,   3.,   3.,   3.,   3.,   3.,   3.,   3.,   3.,
        3.,   3.,   3.,   3.,   3.,   9.,   9.,   9.,   9.,   9.,   9.,
        9.,   9.,   9.,  15.]
    ])

visible_middle = np.array(
                 [1, 1, 1, 1, 1, 1, 1, 1, 1, 
                  1, 1, 1, 1, 1, 1, 1, 1, 1, 
                  1, 1, 1, 1, 1, 1, 1, 1, 1, 
                  1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                 dtype=bool)
visible_east = np.array(
               [1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 
                1, 1, 0, 1, 0, 1, 1, 1, 1, 1, 
                1, 0, 1, 1, 0, 1, 1, 1, 1, 1, 
                0, 1, 0, 1, 1, 1, 1],
                 dtype=bool)
visible_bottomCorner = np.array(
                       [1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 
                        1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 
                        1, 1, 0, 1, 1, 1, 1, 0, 1, 0, 
                        1, 1, 1, 1, 1, 1, 1],
                 dtype=bool)
visible_outside = np.array(
                  [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
                   0, 0, 0, 0],
                 dtype=bool)


class TestVisibilityShoeBox3D(unittest.TestCase):

    def test_sources(self):
        self.assertTrue(np.allclose(images_found, sources))

    def test_visibility_middle(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[0,:]], sources[:,visible_middle]))

    def test_visibility_east(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[1,:]], sources[:,visible_east]))
        
    def test_visibility_southWestBottomCorner(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[2,:]], sources[:,visible_bottomCorner]))
        
    def test_visibility_outside(self):
        self.assertTrue(np.allclose(images_found[:,visibilities[3,:]], sources[:,visible_outside]))

if __name__ == '__main__':
    unittest.main()




from __future__ import division, print_function

import numpy as np
import os
from scipy.stats import multivariate_normal
import sys
import struct

try:
    import sounddevice as sd
    have_sounddevice = True
except:
    have_sounddevice = False

from .acoustics import mfcc

class CircularGaussianEmission:

    def __init__(self, nstates, odim=1, examples=None):
        ''' Initialize the Gaussian emission object '''

        # The emissions parameters
        self.K = nstates

        if examples is None:
            
            # Initialize to random components
            self.O = odim
            self.mu = np.random.normal(size=(self.K, self.O))
            self.Sigma = np.ones((self.K, self.O))*10

        else:
            # Initialize all components to the same mean and variance of the data
            self.O = examples[0].shape[1]

            X = np.concatenate(examples, axis=0)

            self.mu = np.array([np.mean(X, axis=0)]*self.K)
            centered = X - self.mu[0]
            self.Sigma = np.array([np.mean(centered**2, axis=0)]*self.K)

    def update_parameters(self, examples, gamma):

        g = np.concatenate(gamma, axis=0)
        X = np.concatenate(examples, axis=0)
        Z = g.sum(axis=0)

        for k in range(self.K):
            self.mu[k] = np.sum(X.T * g[:,k], axis=1)/Z[k]
            centered = (X - self.mu[k])**2
            self.Sigma[k] = np.sum(centered.T * g[:,k], axis=1)/Z[k]


    def get_pdfs(self):
        ''' Return the pdf of all the emission probabilities '''
        return [multivariate_normal(self.mu[k], np.diag(self.Sigma[k])) for k in range(self.K)]


    def prob_x_given_state(self, examples):
        ''' 
        Recompute the probability of the observation given the state of the
        latent variables
        '''
        distribution = [multivariate_normal(self.mu[k], np.diag(self.Sigma[k])) for k in range(self.K)]
        p_x_given_z = []

        for X in examples:
            p_x_given_z.append(np.zeros((X.shape[0], self.K)))
            for k in range(self.K):
                p_x_given_z[-1][:,k] = distribution[k].pdf(X)

        return p_x_given_z


class GaussianEmission:

    def __init__(self, nstates, odim=1, examples=None):
        ''' Initialize the Gaussian emission object '''

        # The emissions parameters
        self.K = nstates

        if examples is None:
            # initialize to random mean unit variance
            self.O = odim
            self.mu = np.random.normal(size=(self.K, self.O))
            self.Sigma = np.random.normal(size=(self.K, self.O, self.O))
            for k in range(self.K):
                self.Sigma[k] = np.dot(self.Sigma[k].T, self.Sigma[k]) + np.eye(self.O)

        else:
            # Initialize using mean and covariance of dataset
            self.O = examples[0].shape[1]

            X = np.concatenate(examples, axis=0)

            self.mu = np.array([np.mean(X, axis=0)]*self.K)
            centered = X - self.mu[0]
            self.Sigma = np.array([np.diag(np.mean(centered**2, axis=0))]*self.K)

    def update_parameters(self, examples, gamma):

        g = np.concatenate(gamma, axis=0)
        X = np.concatenate(examples, axis=0)
        Z = g.sum(axis=0)

        for k in range(self.K):
            self.mu[k] = np.sum(X.T * g[:,k], axis=1)/Z[k]
            centered = X - self.mu[k]
            self.Sigma[k] = np.dot(centered.T*g[:,k], centered/Z[k])


    def get_pdfs(self):
        ''' Return the pdf of all the emission probabilities '''
        return [multivariate_normal(self.mu[k], self.Sigma[k]) for k in range(self.K)]


    def prob_x_given_state(self, examples):
        ''' 
        Recompute the probability of the observation given the state of the
        latent variables
        '''
        distribution = [ multivariate_normal(self.mu[k], self.Sigma[k]) for k in range(self.K)]
        p_x_given_z = []

        for X in examples:
            p_x_given_z.append(np.zeros((X.shape[0], self.K)))
            for k in range(self.K):
                p_x_given_z[-1][:,k] = distribution[k].pdf(X)

        return p_x_given_z


class HMM:
    '''
    Hidden Markov Model with Gaussian emissions

    Attributes
    ----------
    K : int
        Number of states in the model
    O : int
        Number of dimensions of the Gaussian emission distribution
    A : ndarray
        KxK transition matrix of the Markov chain
    pi : ndarray
        K dim vector of the initial probabilities of the Markov chain
    emission : (GaussianEmission or CircularGaussianEmission)
        An instance of emission_class
    model : string, optional
        The model used for the chain, can be 'full' or 'left-right'
    leftright_jum_max : int, optional
        The number of non-zero upper diagonals in a 'left-right' model
    '''

    def __init__(self, nstates, emission, model='full', leftright_jump_max=3):
        '''
        Initialize a Hidden Markov Model with nstates and Gaussian observations 
        
        nstates: int
            The number of states in the Markov chain
        emission : emission object, optional
            The emission object (CircularGaussianEmission or GaussianEmission)
        model : string, optional
            The model used for the chain, can be 'full' or 'left-right'
        leftright_jump_max : int
            The maximum jump length in the Left-Right chain model
        '''

        self.K = nstates            # number of states
        self.emission = emission    # The observation parameters

        # The Markov chain parameters
        self.model = model

        self.leftright_jump_max = leftright_jump_max
        self.A = np.zeros((self.K, self.K)) # the state transition matrix
        self.pi = np.zeros((self.K))        # the initial distribution

        # Initialize the HMM parameters to some random values
        if self.model == 'full':
            self.A = np.random.uniform(size=(self.K,self.K))
            self.pi = np.random.uniform(size=(self.K))

        elif self.model == 'left-right':
            self.A = np.triu(np.tril(np.random.uniform(size=(self.K,self.K)), k=self.leftright_jump_max))
            self.A += np.diag(np.sum(self.A[:,:], axis=1)*2)
            self.pi = np.zeros(self.K)
            self.pi[0] = 1

        # Normalize the distributions
        for row in self.A:
            row /= row.sum()
        self.pi /= self.pi.sum()

    def fit(self, examples, tol=0.1, max_iter=10, verbose=False):
        '''
        Training of the HMM using the EM algorithm

        Parameters
        ----------
        examples : (list)
            A list of examples used to train the model. Each example is
            an array of feature vectors, each row is a feature vector,
            the sequence runs on axis 0
        tol : (float)
            The training stops when the progress between to steps is less than
            this number (default 0.1)
        max_iter : (int)
            Alternatively the algorithm stops when a maximum number of
            iterations is reached (default 10)
        verbose : bool, optional
            When True, prints extra information about convergence
        '''

        # Make sure to normalize parameters that should be...
        for row in self.A:
            row[:] /= row.sum()
        self.pi[:] /= self.pi.sum()

        # Run the EM algorithm
        loglikelihood_old = -np.inf # log-likelihood
        n_iter = 0
        while True:

            # Initialize new parameters value for accumulation
            loglikelihood = 0.

            # We need to run the forward/backward algorithm for each example and
            # and combine the result to form the new estimates
            gamma = []
            xhi = []
            p_x_given_z = self.emission.prob_x_given_state(examples)

            # Expectation-step
            #-----------------

            for X,pxz in zip(examples, p_x_given_z):

                # check dimension of emission
                if X.shape[1] != self.emission.O:
                    raise ValueError("Error: Emission vectors of all examples should have the same size")

                # First compute alpha and beta using forward/backward algo
                alpha, c = self.forward(X, pxz)
                beta = self.backward(X, pxz, c)

                # Recompute the likelihood of the sequence
                # (Bishop 13.63)
                loglikelihood += np.sum(np.log(c))

                # Now the more interesting quantities
                # gamma(z_n) = p(z_n | X, theta_old) 
                # xhi(z_{n-1}, z_n) = p(z_{n-1}, z_n | X, theta_old)
                gamma.append(alpha * beta)
                xhi.append(np.zeros((X.shape[0]-1, self.K, self.K)))
                for n in range(1,X.shape[0]):
                    xhi[-1][n-1] = np.outer(alpha[n-1], beta[n]*pxz[n])*self.A/c[n]

            # Maximization-step
            #------------------

            # update the Markov Chain parameters
            self.update_parameters(examples, gamma, xhi)

            # Update the emission distribution parameters
            self.emission.update_parameters(examples, gamma)

            # Now check for convergence
            #--------------------------
            n_iter += 1
            epsilon = loglikelihood - loglikelihood_old
            if verbose:
                print('Iterations:', n_iter, 'epsilon:', epsilon, 'LL_new:', loglikelihood)

            # some checks here
            if epsilon < tol:
                if verbose:
                    print('Tolerance reached: stopping.')
                break
            if  n_iter == max_iter:
                if verbose:
                    print('Maximum iterations reached: stopping.')
                break

            loglikelihood_old = loglikelihood

        # return the number of iterations performed
        return n_iter


    def update_parameters(self, examples, gamma, xhi):
        ''' Update the parameters of the Markov Chain '''

        X = np.concatenate(examples, axis=0)
        x = np.concatenate(xhi, axis=0)

        self.pi[:] = np.sum([g[0,:] for g in gamma], axis=0)
        self.A = x.sum(axis=0)

        # normalize to enforce distribution constraints
        self.pi /= np.sum(self.pi)
        for k in range(self.K):
            den = np.sum(self.A[k,:])
            if den < 1e-15:
                self.A[k,:] = 0.
            else:
                self.A[k,:] /= den


    def generate(self, N):
        ''' Generate a random sample of length N using the model '''
        X = np.zeros((N, self.emission.O))
        distributions = self.emission.get_pdfs()

        # pick initial state
        state = np.random.choice(self.K, p=self.pi)

        # now run the chain
        for n in range(0,N):
            # produce emission vector according to current state
            X[n,:] = distributions[state].rvs()
            # pick next state
            state = np.random.choice(self.K, p=self.A[state,:])

        return X


    def loglikelihood(self, X):
        '''
        Compute the log-likelihood of a sample vector using the sum-product algorithm
        '''
        p_x_given_z = self.emission.prob_x_given_state([X])[0]
        alpha, c = self.forward(X, p_x_given_z)

        return np.sum(np.log(c))

    def forward(self, X, p_x_given_z):
        ''' The forward recursion for HMM as described in Bishop Ch. 13 '''

        # initialize the alpha vector
        alpha = np.zeros((X.shape[0], self.K))
        c = np.zeros(X.shape[0])

        # initialize the recursion as
        # p(X | z_k) pi_k
        alpha[0] = p_x_given_z[0]*self.pi
        c[0] = np.sum(alpha[0])
        alpha[0] /= c[0]

        # Run the forward recursion
        for n in range(1,X.shape[0]):
            alpha[n] = p_x_given_z[n]*np.dot(self.A.T, alpha[n-1])
            c[n] = np.sum(alpha[n])
            alpha[n] /= c[n]

        return alpha, c

    def backward(self, X, p_x_given_z, c):
        ''' The backward recursion for HMM as described in Bishop Ch. 13 '''

        # intialize the beta vectors
        beta = np.zeros((X.shape[0], self.K))

        # initialize the recursion
        beta[-1,:] = 1

        # Run the backward recursion
        for n in range(X.shape[0]-2,-1,-1):
            beta[n] = np.dot(self.A, p_x_given_z[n+1]*beta[n+1])/c[n+1]

        return beta


    def viterbi(self):
        x=1

